#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : data.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import os
import sys
import numpy as np
import torch as th
import pickle as pkl
from torchbox import savemat, saveh5, loadmat, loadh5


def sarstore(sardata, params, file):
    """Read SAR file

    Store SAR data and platform to a file (``.pkl`` , ``.h5`` or ``.mat``)

    Args:
        sardata (dict): The SAR raw data.

        params (dict): The SAR platform for obtain the SAR data.
        file (str): Description

    Returns:
        file (str): The file path for storing SAR data and platform
    """

    if np.iscomplex(sardata['raw']).any():
        sardata['raw'] = np.stack((sardata['raw'].real, sardata['raw'].imag), dim=-1)

    folder, filename = os.path.split(file)
    _, ext = os.path.splitext(filename)
    if os.path.exists(folder) is False:
        os.makedirs(folder)

    if ext == '.pkl':
        f = open(file, 'wb')
        pkl.dump(sardata, f, 0)
        pkl.dump(params, f, 0)
        f.close()
    elif ext == '.mat':
        savemat(file, {'sardata': sardata, 'params': params})
    elif ext in ['.hdf5', '.h5']:
        saveh5(file, {'sardata': sardata, 'params': params})

    return 0


def sarread(file):
    """Read SAR file

    Read SAR file (``.pkl`` or ``.mat``) and returns SAR data and platform

    Args:
        file (str): SAR data (contains platform) file path.

    Returns:
        sardata (Instance SarData): The SAR raw data.

        params (Instance of SarPlat): The SAR platform for obtain the SAR data.
    """

    filename, EXT = os.path.splitext(file)

    if EXT == '.pkl':
        f = open(file, 'rb')
        # for python2
        if sys.version_info < (3, 1):
            sardata = pkl.load(f)
            params = pkl.load(f)
        # for python3
        else:
            sardata = pkl.load(f, encoding='latin1')
            params = pkl.load(f, encoding='latin1')
        f.close()
    elif EXT == '.mat':
        data = loadmat(file)
        sardata = data['sardata']
        params = data['params']
        del data

    elif EXT in ['.hdf5', '.h5']:
        data = loadh5(file)
        sardata = data['sardata']
        params = data['params']
        del data
    else:
        raise(TypeError("Not supported! Only support: (pkl, mat, hdf5)!"))

    return sardata, params


def format_data(X, modefrom='chnl_last', modeto='chnl_first'):
    """format data

    Format data to chanel first or chanel last.

    Args:
        X (numpy array): data to be formated
        modefrom (str, optional): chnl_last  --> chanel last; chnl_first  --> chanel first
            (the default is 'chnl_last', which chanel first)
        modeto (str, optional): chnl_last  --> chanel last; chnl_first  --> chanel first
            (the default is 'chnl_last', which chanel first)

    Returns:
        X (numpy array): Formated data

    Raises:
        TypeError: X should be a 3 or 4 dimention array!
        ValueError: Unknown mode of modefrom or modeto
    """

    if modefrom == modeto:
        return X
    if X.dim() == 4:
        s = 1
        m = 2
        e = 3
    elif X.dim() == 3:
        s = 0
        m = 1
        e = 2
    else:
        raise TypeError("X should be a 3 or 4 dimention array!")
    if modefrom == 'chnl_last':
        if modeto == 'chnl_first':
            X = th.transpose(X, s, e)  # N H W C --> N C W H
            X = th.transpose(X, m, e)  # N C W H --> N C H W
        else:
            raise ValueError("Unknown mode of: ", modeto)
    elif modefrom == 'chnl_first':
        if modeto == 'chnl_last':
            X = th.transpose(X, s, e)  # N C H W --> N W H C
            X = th.transpose(X, s, m)  # N W H C --> N H W C
    return X



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : ceos.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import re
import copy
import struct
import logging
import numpy as np
from tqdm import tqdm

from torchsar.products.utils import getnumber, splitfmt
from torchsar.products.record import readrcd, readrcd1item, printrcd


# B: number in binary format
# A: string
# I: char
SarDataFileFileDescriptorRecordCEOS = {
    # SAR DATA FILE - FILE DESCRIPTOR RECORD (FIXED SEGMENT)
    'Record sequence number': [(1, 4), '1B4', 0],

}

SarDataFileSignalDataRecordCEOS = {
    'Record sequence number': [(1, 4), '1B4', 0],

}

SarDataFileProcessedDataRecordCEOS = {
    'Record sequence number': [(1, 4), '1B4', 0],

}


def decfmtfceos(F, n, x, b, e='<'):
    r"""decode format descriptor of CEOS

    decode format descriptor of CEOS.

    ``'nFx'`` --> decode :attr:`b` x Bytes-by-x Bytes as format :attr:`F` specified,
    and do :attr:`n` times.

    Parameters
    ----------
    F : str
        formation string, ``C`` --> Complex(real, imag), ``F`` --> float, ``B`` --> Bytes, ``A`` --> String, ``I`` --> asicc
    n : int
        amount of elements with formation :attr:`F`
    x : int
        number of Bytes
    b : bytes
        bytes to be decoded
    e : str, optional
        endian, ``'<'`` --> little, ``'>'`` -- > big (the default is '<', which means little endian)

    Returns
    -------
    list
        decoded result list.
    """

    raise(ImportError('Not opened yet!'))


def _getdtype_component(fmtrcd, fmtuser):

    raise(ImportError('Not opened yet!'))


def read_ceos_sar_raw(filepath, sl=1, el=-1, rmbp=False):
    r"""read CEOS SAR raw data

    read CEOS SAR raw data pulse from line :attr:`sl` to line :attr:`el`.

    Parameters
    ----------
    filepath : str
        SAR raw data file path string, for ERS1/2 --> ``*.raw`` for RADARSAT --> ``DAT_01.001``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array


    """

    logging.info("===In read_ceos_sar_raw...")

    raise(ImportError('Not opened yet!'))


def read_ceos_sar_slc(filepath, sl=1, el=-1, rmbp=False):
    r"""read CEOS SAR slc data

    read CEOS SAR slc data from line :attr:`sl` to line :attr:`el`.

    Parameters
    ----------
    filepath : str
        SAR slc data file path string, for ERS1/2 --> ``*.D`` for RADARSAT --> ``DAT_01.001``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       Processed SLC SAR data matrix.

    """

    logging.info("===In read_ceos_sar_slc...")

    raise(ImportError('Not opened yet!'))



if __name__ == '__main__':
    from torchsar import decfmtfers
    disk = '/mnt/d/'
    filename = 'E2_84686_STD_L0_F203'
    # filename = 'E2_84699_STD_L0_F303'
    # filepath = disk + 'DataSets/sar/ERS/data/' + filename + '/' + filename + '.000.ldr'
    filepath = disk + 'DataSets/sar/ERS/data/' + \
        filename + '/' + filename + '.000.raw'

    endian = '>'

    D = copy.deepcopy(SarDataFileFileDescriptorRecordCEOS)
    # D = copy.deepcopy(SarDataFileSignalDataRecordERS)

    offset = 0
    # offset = 11644

    readrcd(filepath, decfmtfers, D, offset=offset, endian='>')

    print("===============")
    printrcd(D)



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : range_doppler_algorithm.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import time
import torch as th
import torchbox as tb
import torchsar as ts


class LRDAnet(th.nn.Module):

    def __init__(self, Na, Nr, Fsa, Fsr, Ta, Tp, Ka, Kr, fdc, Fc, R0, Vr, La):
        super(LRDAnet, self).__init__()
        self.Na = Na
        self.Nr = Nr
        self.Fsa = Fsa
        self.Fsr = Fsr
        self.Ta = Ta
        self.Tp = Tp
        self.Ka = Ka
        self.Kr = Kr
        self.fdc = fdc
        self.Fc = Fc
        self.R0 = R0
        self.Vr = Vr
        self.La = La

        Noff, Wl = th.linspace(0, Nr, Nr), tb.C / Fc
        Rp = ts.min_slant_range(R0, Fsr, Noff)

        if Ka is None:
            Ka = ts.dre_geo(Wl, Vr, Rp)

        if Ta is None:
            FPa = ts.azimuth_footprint(Rp, Wl, La)
            Ta = th.mean(FPa).item() / Vr

        raise(ImportError('Not opened yet!'))


    def forward(self, X):
        # X --> Na-Nr-2

        raise(ImportError('Not opened yet!'))


    def focus(self, X, pa=None, pr=None):
        # X --> N-1-Na-Nr-2
        # pa --> N-Na
        # pr --> N-Nr

        if pa is None and pr is None:
            return X

        if pa is not None:
            X = tb.fft(X, nfft=None, axis=2, norm=False)
            pa = pa.reshape(pa.size(0), 1, int(pa.numel() / pa.size(0)), 1)
            epa = th.stack((th.cos(pa), th.sin(pa)), dim=-1)
            X = tb.ematmul(X, epa)
        if pr is not None:
            X = tb.fft(X, nfft=None, axis=3, norm=False)
            pr = pr.reshape(pr.size(0), 1, 1, int(pr.numel() / pr.size(0)))
            epr = th.stack((th.cos(pr), th.sin(pr)), dim=-1)
            X = tb.ematmul(X, epr)
        if pa is not None:
            X = tb.ifft(X, nfft=None, axis=2, norm=False)
        if pr is not None:
            X = tb.ifft(X, nfft=None, axis=3, norm=False)

        return X


if __name__ == "__main__":

    epochs = 10
    learning_rate = 1e-3
    device = th.device('cuda:1' if th.cuda.is_available() else 'cpu')
    # device = th.device('cpu')
    print(device)

    Na, Nr = 512, 512
    net = LRDAnet(Na, Nr)
    net.to(device=device)

    X = th.randn((4, Na, Nr, 2), requires_grad=False)
    X = X.to(device)

    loss_func = tb.Entropy(mode='shannon', cdim=-1, reduction='mean')
    optimizer = th.optim.Adam(net.parameters(), lr=learning_rate)

    for k in range(epochs):
        tstart = time.time()

        pa, Y = net.forward(X)
        loss = loss_func(Y)

        optimizer.zero_grad()

        # Backward pass: compute gradient of the loss with respect to model parameters
        loss.backward()

        # Calling the step function on an Optimizer makes an update to its parameters
        optimizer.step()

        tend = time.time()

        print("---loss: %s, time: %ss" % (loss.item(), tend - tstart))



def mcaf_sm(x, phi=None, niter=10, tol=1e-4, eta=0.1, method='N-MCA', selscat=False, axis=-2, ftshift=True, islog=False):
    r"""Contrast based autofocus

    Maximum-Contrast based Autofocus (MCA)

    Parameters
    ----------
    x : Tensor
        complex image with shape :math:`N×N_a×N_r` or :math:`N×N_a×N_r×2`
    phi : float, optional
        initial value of :math:`\phi` (the default is None, which means zeros)
    niter : int, optional
        number of iterations (the default is 10)
    method : str, optional
        method used to update the phase error
        - ``'FP-MCA'`` --> Fix Point
        - ``'CD-MCA'`` --> Coordinate Descent
        - ``'SU-MCA'`` --> Simultaneous Update
        - ``'N-MCA'`` --> Newton, see [2], [1] has problem when used to small image
        - ``'SN-MCA'`` --> Simplified Newton
        - ``'MN-MCA'`` --> Modified Newton
    selscat : bool, optional
        Select brighter scatters.
    axis : int, optional
        Compensation axis.
    ftshift :  bool, optional
        Does shift zero frequency to center?
    islog :  bool, optional
        Does print log info.


    see [1] Zhang S , Liu Y , Li X . Fast Entropy Minimization Based Autofocusing Technique for ISAR Imaging[J]. IEEE Transactions on Signal Processing, 2015, 63(13):3425-3434.
        [2] Zeng T , Wang R , Li F . SAR Image Autofocus Utilizing Minimum-Entropy Criterion[J]. IEEE Geoence & Remote Sensing Letters, 2013, 10(6):1552-1556.
    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : focusing.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import torch as th
from torch.nn.parameter import Parameter
from torchsar.autofocus.focusing import focus


class AutoFocus(th.nn.Module):

    def __init__(self, Na, Nr, pa=None, pr=None, ftshift=False, trainable=True):
        super(AutoFocus, self).__init__()

        self.Na = Na
        self.Nr = Nr
        self.ftshift = ftshift

        if Na is not None:
            if pa is None:
                pa = th.zeros(1, Na)
            self.pa = Parameter(pa, requires_grad=trainable)

        if Nr is not None:
            if pr is None:
                pr = th.zeros(1, Nr)
            self.pr = Parameter(pr, requires_grad=trainable)

    def forward(self, X, isfft=True):
        X = focus(X, pa=self.pa, pr=self.pr, isfft=isfft, ftshift=self.ftshift)

        return X



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : utils.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import re


def getnumber(b):
    r"""obtain number in string or bytes

    obtain number in string/bytes, e.g. b' 123 45 ' --> [123, 45],
    b' 123, 45 ' --> [123, 45] and b'  ' --> []

    Parameters
    ----------
    b : bytes
        bytes string for extracting numbers.

    Returns
    -------
    n : number list or number
        extracted number list, if there only one element, return a number.
    """
    n = [int(i) for i in re.findall(r'\d+', str(b))]
    if len(n) == 1:
        n = n[0]
    return n


def splitfmt(fmt):
    n, F, x = re.findall(r'[0-9]+|[a-z,A-Z]+|[0-9]', fmt)

    return (int(n), F, int(x))


if __name__ == '__main__':

    print(getnumber(b' 1 23'))
    print(getnumber(b' 1,23'))
    print(getnumber(b'  '))

    print(splitfmt('100IU33'))



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : record.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import numpy as np
import re
from .utils import splitfmt


def readrcd(filepath, decfmtf, D, offset=0, endian='>'):
    """read file record

    read record information from a file. see :func:`read_ers_sar_raw` for reading ERS SAR raw data.

    Parameters
    ----------
    filepath : str
        file path string
    decfmtf : function
        format decoding function. see :func:`decfmtfers` for example.
    D : dict
        record descriptor dict, each value will be rewrited after reading.
    offset : int, optional
        record offset, read from the offset-th Byte (the default is 0)
    endian : str, optional
        endian, ``'<'`` --> little, ``'>'`` -- > big (the default is '<', which means little endian)

    Returns
    -------
    state : number
        reading status. 0 --> OK and done; 1 --> no such file; 2 --> bad record

    """

    state = 0

    raise(ImportError('Not opened yet!'))



def readrcd1item(filepath, decfmtf, adrfmt, offset=0, endian='>'):
    """read one file record item

    read one record item information from a file. see :func:`read_ceos_sar_raw` for reading ERS SAR raw data.

    Parameters
    ----------
    filepath : str
        file path string
    decfmtf : function
        format decoding function. see :func:`decfmtfers` for example.
    adrfmt : tuple
        address and formation of one record item, a tuple with two
        elements: ``(adr, fmt)``, where, ``adr`` = (start address, end address) is a tuple and specifies
        the address (unit Bytes), ``fmt`` is the formation string.
    offset : int, optional
        record address offset, begin from the offset-th Byte (the default is 0)
    endian : str, optional
        endian, ``'<'`` --> little, ``'>'`` -- > big (the default is '<', which means little endian)

    Returns
    -------
    v : {}
        readed record item value.

    """

    raise(ImportError('Not opened yet!'))



def printrcd(D):
    """print record

    print record in ascend address order.

    Parameters
    ----------
    D : dict
        descriptor dict
    """

    for k, v in sorted(D.items(), key=lambda item: item[1][0]):
        print(k, v)



def azimuth_beamwidth(Wl, La):
    ...

def azimuth_footprint(R, Wl, La):
    ...

def compute_range_beamwidth2(Nr, Fsr, H, Aon):
    r"""computes beam angle in range direction


    Parameters
    ----------
    Nr : int
        Number of samples in range direction.
    Fsr : float
        Sampling rate in range direction.
    H : float
        Height of the platform.
    Aon : float
        The off-nadir angle (unit: rad).

    Returns
    -------
    float
        The beam angle (unit, rad) in range direction.

    """

def cr_footprint(Wl, H, La, Ad):
    r"""cross range (azimuth) foot print

    .. math::
       R_{CR} \approx \frac{\lambda}{L_a}\frac{H}{{\rm cos}\theta_d}

    Parameters
    ----------
    Wl : float
        wave length
    H : float
        height of SAR platform
    La : float
        length of antenna aperture (azimuth)
    Ad : float
        depression angle

    Returns
    -------
    float
        foot print size in azimuth
    """





def _getdtype_component(fmtrcd, fmtuser):
    ...

def get_ers_sar_plat_position(D):
    """get platform position data

    return a numpy array
        1st: X Y Z Vx Vy Vz
        2st: X Y Z Vx Vy Vz
           :
        nst: X Y Z Vx Vy Vz

    Parameters
    ----------
    D : dict
        LeaderFileImportantImagingParametersRecordALOS dict
    """

def read_ers_sar_ldr_iip(filepath, verbose=False):
    r"""Read important imaging parameters from leader file

    Read important imaging parameters from leader file.

    Parameters
    ----------
    filepath : str
        Leader file path string.
    """

def read_ers_sar_raw(rawfile, ledfile, sl=1, el=-1, rmbp=False):
    r"""read ERS SAR raw data

    read ERS1/2 SAR raw data pulse from line :attr:`sl` to line :attr:`el`.


    Parameters
    ----------
    rawfile : str
        ERS SAR raw data file path string, for ERS1/2 --> ``*.raw``
        e.g. ``E2_81988_STD_L0_F327.000.raw``
    ledfile : str
        ERS SAR leader file path string, ---> ``*.ldr``, e.g. ``E2_81988_STD_L0_F327.000.ldr``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       SAR raw signal data matrix.

    Pd : dict
       Parameter dictionary.

    """

def read_ers_sar_slc(filepath, sl=1, el=-1, rmbp=False):
    r"""read ERS SAR SLC data

    read ERS1/2 SAR SLC data pulse from line :attr:`sl` to line :attr:`el`.


    Parameters
    ----------
    filepath : str
        ERS SAR raw data file path string, for ERS1/2 --> ``*.raw``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       SAR SLC data matrix.

    """





def focus(x, pa=None, pr=None, isfft=True, ftshift=True):
    r"""Focus image with given phase error

    Focus image in azimuth by

    .. math::
        Y(k, n_r)=\sum_{n_a=0}^{N_a-1} X(n_a, n_r) \exp \left(-j \varphi_{n_a}\right) \exp \left(-j \frac{2 \pi}{N_a} k n_a\right)

    where, :math:`\varphi_{n_a}` is the estimated azimuth phase error of the :math:`n_a`-th azimuth line, :math:`y(k, n_r)` is
    the focused pixel.

    The focus method in range is the same as azimuth.

    Args:
        x (Tensor): Defocused image data :math:`{\mathbf X} \in{\mathbb C}^{N\times N_c\times N_a\times N_r}` or
            :math:`{\mathbf X} \in{\mathbb R}^{N\times N_a\times N_r\times 2}` or
            :math:`{\mathbf X} \in{\mathbb R}^{N\times N_c\times N_a\times N_r\times 2}`.
        pa (Tensor, optional): Focus parameters in azimuth, phase error in rad unit. (the default is None, not focus)
        pr (Tensor, optional): Focus parameters in range, phase error in rad unit. (the default is None, not focus)
        isfft (bool, optional): Is need do fft (the default is True)
        ftshift (bool, optional): Is shift zero frequency to center when do fft/ifft/fftfreq (the default is True)

    Returns:
        (Tensor): A tensor of focused images.

    Raises:
        TypeError: :attr:`x` is complex and should be in complex or real represent formation!
    """

def defocus(x, pa=None, pr=None, isfft=True, ftshift=True):
    r"""Defocus image with given phase error

    Defocus image in azimuth by

    .. math::
        Y(k, n_r)=\sum_{n_a=0}^{N_a-1} X(n_a, n_r) \exp \left(j \varphi_{n_a}\right) \exp \left(-j \frac{2 \pi}{N_a} k n_a\right)

    where, :math:`\varphi_{n_a}` is the estimated azimuth phase error of the :math:`n_a`-th azimuth line, :math:`y(k, n_r)` is
    the focused pixel.

    The defocus method in range is the same as azimuth.

    Args:
        x (Tensor): Focused image data :math:`{\mathbf X} \in{\mathbb C}^{N\times N_c\times N_a\times N_r}` or
            :math:`{\mathbf X} \in{\mathbb R}^{N\times N_a\times N_r\times 2}` or
            :math:`{\mathbf X} \in{\mathbb R}^{N\times N_c\times N_a\times N_r\times 2}`.
        pa (Tensor, optional): Defocus parameters in azimuth, phase error in rad unit. (the default is None, not focus)
        pr (Tensor, optional): Defocus parameters in range, phase error in rad unit. (the default is None, not focus)
        isfft (bool, optional): Is need do fft (the default is True)
        ftshift (bool, optional): Is shift zero frequency to center when do fft/ifft/fftfreq (the default is True)

    Returns:
        (Tensor): A tensor of defocused images.

    Raises:
        TypeError: :attr:`x` is complex and should be in complex or real represent formation!
    """





def rcmc_interp(Sr, tr, D):
    r"""range migration correction with linear interpolation

    Range migration correction with linear interpolation.

    Parameters
    ----------
    Sr : tensor
        SAR raw data :math:`N_a×N_r` in range dopplor domain
    tr : 1d-tensor
        time array :math:`N_r×1` in range
    D : 1d-tensor
        :math:`N_a×1` migration factor vector

    Returns
    -------
    Srrcmc
        data after range migration correction :math:`N_a×N_r`
    """





def getnumber(b):
    r"""obtain number in string or bytes

    obtain number in string/bytes, e.g. b' 123 45 ' --> [123, 45],
    b' 123, 45 ' --> [123, 45] and b'  ' --> []

    Parameters
    ----------
    b : bytes
        bytes string for extracting numbers.

    Returns
    -------
    n : number list or number
        extracted number list, if there only one element, return a number.
    """

def splitfmt(fmt):




def _getdtype_component(fmtrcd, fmtuser):
    ...

def get_alos_palsar_plat_position(D):
    """get platform position data

    return a numpy array
        1st: X Y Z Vx Vy Vz
        2st: X Y Z Vx Vy Vz
           :
        nst: X Y Z Vx Vy Vz

    Parameters
    ----------
    D : dict
        LeaderFileImportantImagingParametersRecordALOS dict
    """

def get_alos_palsar_attitude(D):
    """get attitude data

    return a numpy array
        1st: Pitch Roll Yaw PitchRate RollRate YawRate
        2st: Pitch Roll Yaw PitchRate RollRate YawRate
           :
        nst: Pitch Roll Yaw PitchRate RollRate YawRate

    Parameters
    ----------
    D : dict
        LeaderFileImportantImagingParametersRecordALOS dict
    """

def read_alos_palsar_ldr_iip(ledfile, verbose=False):
    r"""Read important imaging parameters from leader file

    Read important imaging parameters from leader file.

    For example, ``LED-ALPSRP050500980-H1.0__A``

    Parameters
    ----------
    ledfile : str
        Leader file path string.
    """

def read_alos_palsar_raw(rawfile, ledfile, sl=1, el=-1, rmbp=False):
    r"""read ALOS PALSAR raw data

    read ALOS PALSAR raw data pulse from line :attr:`sl` to line :attr:`el`. This function call
    function :func:`read_ceos_sar_raw` firstly and do some post-process.


    Parameters
    ----------
    rawfile : str
        ALOS PALSAR raw data file path string, ---> ``IMG*.0__A``, e.g. ``IMG-HH-ALPSRP050500980-H1.0__A``
    ledfile : str
        ALOS PALSAR leader file path string, ---> ``LED*.0__A``, e.g. ``LED-ALPSRP050500980-H1.0__A``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    Sr : 2d-array
       SAR raw signal data matrix.

    Pd : dict
       Parameter dictionary.

    """

def read_alos_palsar_slc(filepath, sl=1, el=-1, rmbp=False):
    r"""read ALOS PALSAR SLC data

    read ALOS PALSAR SLC data pulse from line :attr:`sl` to line :attr:`el`. This function call
    function :func:`read_ceos_sar_slc` firstly.


    Parameters
    ----------
    filepath : str
        ALOS PALSAR raw data file path string, for ALOSPALSAR --> ``IMG*.1__A``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       SAR SLC data matrix.

    """





def _getdtype_component(fmtrcd, fmtuser):
    ...

def read_radarsat_sar_raw(filepath, sl=1, el=-1, rmbp=False):
    r"""read RADARSAT SAR raw data

    read RADARSAT raw data pulse from line :attr:`sl` to line :attr:`el`. This function call
    function :func:`read_ceos_sar_raw` firstly and do some post-process.


    Parameters
    ----------
    filepath : str
        RADARSAT SAR raw data file path string with format ``.001``, such as ``DAT_01.001``.
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
        SAR raw data.

    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : autofocus.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import torch as th
import torchbox as tb
import torchsar as ts


def _focusing(X, phi, Na, Nr, ftshift=False, isfft=True):
    if th.is_complex(X):
        # N, Na, Nr = X.size(0), X.size(-2), X.size(-1)
        X = th.view_as_real(X)
        crepflag = True
    elif X.size(-1) == 2:
        # N, Na, Nr = X.size(0), X.size(-3), X.size(-2)
        crepflag = False
    else:
        raise TypeError('X is complex and should be in complex or real represent formation!')

    d = X.dim()
    sizea, sizer = [1] * d, [1] * d
    returns = []
    if Na is not None:
        pa = phi[0]
        sizea[0], sizea[-3], sizea[-1] = pa.size(0), pa.size(1), 2
        epa = th.stack((th.cos(pa), -th.sin(pa)), dim=-1)
        epa = epa.reshape(sizea)

        if isfft:
            X = tb.fft(X, cdim=-1, dim=-3, keepcdim=True, shift=ftshift)
        X = tb.ematmul(X, epa, cdim=-1)
        X = tb.ifft(X, cdim=-1, dim=-3, keepcdim=True, shift=ftshift)
        returns.append(pa.data)

    if Nr is not None:
        if Na is None:
            pr = phi[0]
        else:
            pr = phi[1]
        sizer[0], sizer[-2], sizer[-1] = pr.size(0), pr.size(1), 2
        epr = th.stack((th.cos(pr), -th.sin(pr)), dim=-1)
        epr = epr.reshape(sizer)

        if isfft:
            X = tb.fft(X, cdim=-1, dim=-2, keepcdim=True, shift=ftshift)
        X = tb.ematmul(X, epr, cdim=-1)
        X = tb.ifft(X, cdim=-1, dim=-2, keepcdim=True, shift=ftshift)
        returns.append(pr.data)
    
    if crepflag:
        X = th.view_as_complex(X)

    return [X] + returns


class AutoFocusBarePhi(th.nn.Module):

    def __init__(self, Na, Nr, Nb=None, pa=None, pr=None, Ma=None, Mr=None, xa=None, xr=None, ftshift=False, trainable=True):
        super(AutoFocusBarePhi, self).__init__()

        if type(Ma) is int:
            Ma = (2, Ma)
        if type(Mr) is int:
            Mr = (2, Mr)
        if Nb is None:
            Nb = 1

        self.Na = Na
        self.Nr = Nr
        self.Nb = Nb
        self.Ma = Ma
        self.Mr = Mr
        self.ftshift = ftshift

        self.barephi = ts.BarePhi(Na, Nr, Nb, pa=pa, pr=pr, Ma=Ma, Mr=Mr, xa=xa, xr=xr, shift=ftshift, trainable=trainable)

    def forward(self, X, isfft=True):

        phi = self.barephi()

        returns = _focusing(X, phi, self.Na, self.Nr, ftshift=self.ftshift, isfft=isfft)

        return returns

    def get_param(self):
        param = []
        if self.barephi.pa is not None:
            param.append(self.barephi.pa)
        if self.barephi.pr is not None:
            param.append(self.barephi.pr)
        return param

    def param_init(self, pa=None, pr=None):
        self.barephi.param_init(pa, pr)


class AutoFocusPolyPhi(th.nn.Module):

    def __init__(self, Na, Nr, Nb=None, ca=None, cr=None, Ma=2, Mr=2, xa=None, xr=None, ftshift=False, trainable=True):
        super(AutoFocusPolyPhi, self).__init__()

        if type(Ma) is int:
            Ma = (2, Ma)
        if type(Mr) is int:
            Mr = (2, Mr)
        if Nb is None:
            Nb = 1

        self.Na = Na
        self.Nr = Nr
        self.Nb = Nb
        self.Ma = Ma
        self.Mr = Mr
        self.ftshift = ftshift

        self.polyphi = ts.PolyPhi(Na, Nr, Nb, ca=ca, cr=cr, Ma=Ma, Mr=Mr, xa=xa, xr=xr, shift=ftshift, trainable=trainable)

    def forward(self, X, isfft=True):

        phi = self.polyphi()

        returns = _focusing(X, phi, self.Na, self.Nr, ftshift=self.ftshift, isfft=isfft)

        return returns

    def get_param(self):
        param = []
        if self.polyphi.ca is not None:
            param.append(self.polyphi.ca)
        if self.polyphi.cr is not None:
            param.append(self.polyphi.cr)
        return param

    def param_init(self, ca=None, cr=None):
        self.polyphi.param_init(ca, cr)


class AutoFocusDctPhi(th.nn.Module):

    def __init__(self, Na, Nr, Nb=None, ca=None, cr=None, Pa=4, Pr=4, ftshift=False, trainable=True):
        super(AutoFocusDctPhi, self).__init__()

        if Nb is None:
            Nb = 1

        self.Na = Na
        self.Nr = Nr
        self.Nb = Nb
        self.Pa = Pa
        self.Pr = Pr
        self.ftshift = ftshift

        self.dctphi = ts.DctPhi(Na, Nr, Nb, ca=ca, cr=cr, Pa=Pa, Pr=Pr, shift=ftshift, trainable=trainable)

    def forward(self, X, isfft=True):

        phi = self.dctphi()

        returns = _focusing(X, phi, self.Na, self.Nr, ftshift=self.ftshift, isfft=isfft)

        return returns

    def get_param(self):
        param = []
        if self.dctphi.ca is not None:
            param.append(self.dctphi.ca)
        if self.dctphi.cr is not None:
            param.append(self.dctphi.cr)
        return param

    def param_init(self, ca=None, cr=None):
        self.dctphi.param_init(ca, cr)


class AutoFocusSinPhi(th.nn.Module):

    def __init__(self, Na, Nr, Nb=None, aa=None, fa=None, ar=None, fr=None, ftshift=False, trainable=True):
        super(AutoFocusSinPhi, self).__init__()

        if Nb is None:
            Nb = 1

        self.Na = Na
        self.Nr = Nr
        self.Nb = Nb
        self.ftshift = ftshift

        self.sinphi = ts.SinPhi(Na, Nr, Nb, aa=aa, fa=fa, ar=ar, fr=fr, shift=ftshift, trainable=trainable)

    def forward(self, X, isfft=True):
        d = X.dim()

        phi = self.sinphi()

        returns = _focusing(X, phi, self.Na, self.Nr, ftshift=self.ftshift, isfft=isfft)

        return returns

    def get_param(self):
        param = []
        if self.sinphi.fa is not None:
            param.append(self.sinphi.aa)
            param.append(self.sinphi.fa)
        if self.sinphi.fr is not None:
            param.append(self.sinphi.ar)
            param.append(self.sinphi.fr)
        return param

    def param_init(self, aa=None, fa=None, ar=None, fr=None):
        self.sinphi.param_init(aa, fa, ar, fr)



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : compute_sar_parameters_func.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import math
import torch as th
import numpy as np
from torchbox.utils.const import *

AirboneSatelliteHeightBoundary = 60e3  # 60 km


def compute_sar_parameters(pdict, islog=False):

    trmod = 'WithTp'
    trmod = 'WithoutTp'

    if 'GeometryMode' not in pdict:
        pdict['GeometryMode'] = 'SceneGeometry'
    if pdict['GeometryMode'] is None:
        pdict['GeometryMode'] = 'BeamGeometry'

    if islog:
        print("---In compute_sar_parameters------------------------------------------------------")

    Re = Ree
    Rs = pdict['H'] + Re
    pdict['Rs'] = Rs
    if 'Vr' not in pdict or pdict['Vr'] is None:
        pdict['Vr'] = pdict['V']
    if 'V' not in pdict or pdict['V'] is None:
        pdict['V'] = pdict['Vr']

    if pdict['H'] < AirboneSatelliteHeightBoundary:
        print("---Airbone.")
        pdict['Vs'] = pdict['V']
        pdict['Vr'] = pdict['V']
        pdict['Vg'] = pdict['V']
        pdict['To'] = -1
        pdict['Ws'] = -1
    else:
        # print("---Satellite.")
        # Earth Bending Geometry (Vg < Vr < Vs, Vr = \sqrt{Vg*Vs})
        pdict['To'] = 2 * PI * math.sqrt((Rs)**3 / Ge)
        pdict['Ws'] = 2 * PI / pdict['To']
        pdict['Vs'] = Rs * pdict['Ws']
        pdict['Vr'] = pdict['V']
        # Equivalent velocity (Linear geometry)
        pdict['Vg'] = pdict['Vr']**2 / pdict['Vs']  # ground velocity
    Vr = pdict['Vs'] * math.sqrt(Re / Rs)
    if islog:
        print("---Computed equivalent velocity is %.4f, setted is %.4f (utilized)" %
              (Vr, pdict['Vr']))
    pdict['Vs'] = Vr / math.sqrt(Re / Rs)  # reset

    if 'PlatCenter' not in pdict or pdict['PlatCenter'] is None:
        pdict['PlatCenter'] = [0, 0, pdict['H'], 0, pdict['Vr'], 0]
    if 'SceneCenter' not in pdict or pdict['SceneCenter'] is None:
        pdict['SceneCenter'] = [0, 0, 0]

    # As: squint angle in earth curve geometry, Ar: squint angle in line geometry
    pdict['Ar'] = (pdict['Vs'] / pdict['Vr']) * pdict['As']
    if 'PRF' not in pdict or pdict['PRF'] is None:
        pdict['PRF'] = pdict['Fsa']
    if 'Fsa' not in pdict or pdict['Fsa'] is None:
        pdict['Fsa'] = pdict['PRF']
    if 'Fsr' not in pdict or pdict['Fsr'] is None:
        pdict['Fsr'] = pdict['Fs']
    if 'Fs' not in pdict or pdict['Fs'] is None:
        pdict['Fs'] = pdict['Fsr']

    SceneCenter = pdict['SceneCenter']  # [x, y, z, vx, vy, vz]
    pdict['Wl'] = C / pdict['Fc']
    pdict['Br'] = abs(pdict['Kr'] * pdict['Tp'])
    Na, Nr = pdict['EchoSize'][0:2]

    pdict['Ta'], pdict['Tr'] = (Na * 1.0) / pdict['Fsa'], (Nr * 1.0) / pdict['Fsr']
    pdict['Tsa'], pdict['Tsr'] = 1. / pdict['Fsa'], 1.0 / pdict['Fsr']

    if 'Aon' not in pdict or pdict['Aon'] is None:
        pdict['Aon'] = math.acos(pdict['H'] / pdict['Rnear'])

    pdict['Rnear'] = pdict['H'] / (math.cos(pdict['Aon']) + EPS)
    if trmod == 'WithoutTp':
        pdict['Rfar'] = pdict['Rnear'] + pdict['Tr'] * C / 2
    if trmod == 'WithTp':
        pdict['Rfar'] = pdict['Rnear'] + (pdict['Tr'] - pdict['Tp']) * C / 2
        # pdict['Rfar'] = pdict['Rnear'] + pdict['Tr'] * C / 2

    pdict['tnear'] = pdict['Rnear'] / C
    pdict['tfar'] = pdict['Rfar'] / C
    pdict['SWST'] = 2. * pdict['tnear']  # Sampling Window Start Time, SWST
    pdict['Abr'] = math.acos(pdict['H'] / pdict['Rfar']) - abs(pdict['Aon'])
    pdict['Ad'] = np.sign(pdict['Aon']) * (PI / 2. - (abs(pdict['Aon']) + pdict['Abr'] / 2.))
    pdict['FPr'] = math.sqrt(pdict['Rfar']**2 + pdict['Rnear']**2 - 2. *
                             pdict['Rnear'] * pdict['Rfar'] * math.cos(pdict['Abr']) + EPS)
    # pdict['Rsc'] = 0.5 * math.sqrt(2. * pdict['Rfar']**2 + 2. * pdict['Rnear']**2 - pdict['FPr']**2 + EPS)
    pdict['Rsc'] = math.sqrt(
        pdict['H']**2 + ((pdict['Rfar']**2 - pdict['Rnear']**2) / 2 / pdict['FPr'])**2)

    pdict['Rs0'] = abs(pdict['Rsc'] * math.cos(pdict['Ar']))

    pdict['Ysc'] = pdict['Rsc'] * math.sin(pdict['Ar'])
    pdict['Xsc'] = math.sqrt(abs(pdict['Rs0']**2 - pdict['H']**2))
    # print("[[[[Rs0, Rsc, H, Xsc", pdict['Rs0'], pdict['Rsc'], pdict['H'], pdict['Xsc'])

    pdict['Ynear'] = pdict['Rnear'] * math.sin(pdict['Ar'])
    pdict['Yfar'] = pdict['Rfar'] * math.sin(pdict['Ar'])
    pdict['Xnear'] = math.sqrt(abs((pdict['Rnear']**2 - pdict['H']**2) - pdict['Ynear']**2 + EPS))
    pdict['Xfar'] = math.sqrt(abs((pdict['Rfar']**2 - pdict['H']**2) - pdict['Yfar']**2 + EPS))
    SceneArea = [0, 0, 0, 0]
    SceneArea[0] = pdict['Xnear'] - pdict['Xsc']
    SceneArea[1] = pdict['Xfar'] - pdict['Xsc']
    SceneArea[2] = -pdict['Ta'] * pdict['Vr'] / 2.0
    SceneArea[3] = pdict['Ta'] * pdict['Vr'] / 2.0
    # print("Rnear, Rfar, Xnear, Xfar", pdict['Rnear'], pdict['Rfar'], pdict['Xnear'], pdict['Xfar'], "]]]]]]")
    SceneCenter[0], SceneCenter[1] = pdict['Xsc'], pdict['Ysc']
    pdict['SceneCenter'] = SceneCenter
    pdict['SceneArea'] = SceneArea
    pdict['Na'], pdict['Nr'] = pdict['EchoSize'][0:2]

    pdict['FPr0'] = pdict['H'] / \
        math.tan(pdict['Ad'] + pdict['Abr'] / 2.) - pdict['H'] / math.tan(pdict['Ad'])
    pdict['FPr1'] = pdict['H'] / \
        math.tan(pdict['Ad'] - pdict['Abr'] / 2.) - pdict['H'] / math.tan(pdict['Ad'])
    BeamArea = (pdict['FPr0'], pdict['FPr1'], -pdict['Ta'] *
                pdict['Vg'] / 2.0, pdict['Ta'] * pdict['Vg'] / 2.0)
    pdict['Rbc'] = pdict['H'] / (math.sin(pdict['Ad']) + EPS)
    pdict['Rb0'] = pdict['Rbc'] * math.cos(pdict['Ar'])

    pdict['Ybc'] = pdict['Rbc'] * math.sin(pdict['Ar'])
    pdict['Xbc'] = math.sqrt(abs(pdict['Rb0']**2 - pdict['H']**2))
    pdict['BeamCenter'] = [pdict['Xbc'], pdict['Ybc'], 0.0]
    pdict['BeamArea'] = BeamArea

    if trmod == 'WithoutTp':
        pdict['tstart'] = 2 * pdict['tnear']
        pdict['tend'] = 2 * pdict['tfar']
    if trmod == 'WithTp':
        # tmid = (pdict['Rfar'] + pdict['Rnear']) / C
        pdict['tstart'] = 2 * pdict['tnear']
        pdict['tend'] = 2 * pdict['tfar'] + pdict['Tp']
        # pdict['tstart'] = 2 * pdict['tnear'] - pdict['Tp'] / 2.
        # pdict['tend'] = 2 * pdict['tfar'] + pdict['Tp'] / 2.
        # pdict['tstart'] = tmid - pdict['Tr'] / 2.
        # pdict['tend'] = tmid + pdict['Tr'] / 2.

    # assert (pdict['tend'] - pdict['tstart']) > pdict[
    #     'Tp'], "~~~Range sampling time %.4f(us) too small, should larger than pulse width %.4f(us)" % (pdict['Tr'] * 1e6, pdict['Tp'] * 1e6)

    pdict['ta'] = th.linspace(-pdict['Ta'] / 2., pdict['Ta'] / 2., Na).reshape(Na, 1)
    pdict['tr'] = th.linspace(pdict['tstart'], pdict['tend'], Nr).reshape(1, Nr)

    pdict['Lsar'] = (C / pdict['Fc']) * pdict['Rnear'] / pdict['La']
    pdict['Tsar'] = pdict['Lsar'] / pdict['Vr']
    pdict['Nsar'] = int(pdict['Tsar'] * pdict['Fsa'])

    pdict['Ka'] = (-2.0 * pdict['Vr']**2 * math.cos(pdict['Ar'])**3) / \
        (pdict['Wl'] * pdict['Rnear'])
    pdict['Ba'] = abs(pdict['Ka'] * pdict['Tsar'])

    if islog:
        if pdict['Fsa'] / pdict['Ba'] < 1.1 or pdict['Fsa'] / pdict['Ba'] > 2.0:
            print(
                "---Azimuth sampling rate (Fsa=%f) should be in range [1.1, 2.0]*Ba=[1.1, 2.0]*%f!" % (pdict['Fsa'], pdict['Ba']))
        if pdict['Fsr'] / pdict['Br'] < 1.1 or pdict['Fsr'] / pdict['Br'] > 2.0:
            print(
                "---Range sampling rate (Fsr=%.4e) should be in range [1.1, 2.0]*Br=[1.1, 2.0]*%.4e!" % (pdict['Fsr'], pdict['Br']))

    Fsa = 1. / (pdict['Tr'] + pdict['Tp'])
    assert (pdict['Tsa'] - pdict['Tp']) > pdict['Tr'], "Azimuth sampling rate (Fsa=%.4f) should be lower than %.4f to avoid range ambiguity (1.0 ÷ Fsa - Tp > Tr)" % (pdict['Fsa'], Fsa)

    pdict['osfa'] = pdict['Fsa'] / pdict['Ba']
    pdict['osfr'] = pdict['Fsr'] / pdict['Br']
    print(pdict['Fsa'], pdict['Ba'])
    print(pdict['Fsr'], pdict['Br'])
    assert pdict['osfa'] > pdict['osfr'], "Over sampling factor in azimuth=%.4f should be larger than range=%.4f" % (pdict['osfa'], pdict['osfr'])

    if pdict['GeometryMode'] == 'SceneGeometry':
        pdict['R0'] = pdict['Rs0']

    if pdict['GeometryMode'] == 'BeamGeometry':
        pdict['R0'] = pdict['Rb0']
    pdict['tac'] = -pdict['R0'] * math.tan(pdict['Ar']) / pdict['Vr']

    if 'fdc' not in pdict:
        pdict['fdc'] = 2.0 * pdict['Vr'] * math.sin(pdict['Ar']) / pdict['Wl']

    pdict['Bdop'] = 0.886 * (2 * pdict['Vs'] * math.cos(pdict['As']) / pdict['La'])

    if pdict['H'] < AirboneSatelliteHeightBoundary:
        pdict['Ai'] = PI / 2. - pdict['Ad']
    else:
        pdict['Ao'] = math.acos((Rs**2 + Rea**2 - pdict['Rbc']**2) / (2. * Rs * Rea))
        pdict['Ai'] = (PI / 2. - pdict['Ad']) + pdict['Ao']

    pdict['AzimuthResolution'] = pdict['La'] / 2.
    pdict['SlantRangeResolution'] = C / 2. / pdict['Br']
    pdict['GroundRangeResolution'] = pdict['SlantRangeResolution'] / (math.sin(pdict['Ai']) + EPS)
    pdict['ImagePixelResolution'] = [(pdict['SceneArea'][3] - pdict['SceneArea'][2]) /
                                     pdict['Na'], (pdict['SceneArea'][1] - pdict['SceneArea'][0]) / pdict['Nr']]

    if islog:
        print("---SAR platform height: %.4f(m)" % pdict['H'])
        print("---SAR platform velocity: %.4f(m/s)" % pdict['V'])
        print("---Carrier frequency: %.4f(Hz)" % pdict['Fc'])
        print("---Wavelength: %.4f(m)" % pdict['Wl'])
        print("---Squint angle: %.4f(degree)" % (pdict['As'] * 180 / PI))
        print("---Depression angle: %.4f(degree)" % (pdict['Ad'] * 180 / PI))
        print("---Antenna range beamwidth: %.4f(degree)" % (pdict['Abr'] * 180 / PI))
        print("---off-nadir angle: %.4f(degree)" % (pdict['Aon'] * 180 / PI))
        print("---Azimuth antenna length: %.4f(m)" % pdict['La'])
        print("---Azimuth bandwidth: %.4f(Hz)" % pdict['Ba'])
        print("---Azimuth sampling rate: %.4f(Hz)" % pdict['Fsa'])
        print("---Azimuth over sampling factor: %.4f" % pdict['osfa'])
        print("---Azimuth sampling time: %.4f(s)" % pdict['Ta'])
        print("---Azimuth doppler bandwidth: %.4f(Hz)" % pdict['Bdop'])
        print("---Chirp rate: %.4f(Hz)" % pdict['Kr'])
        print("---Transmitted pulse width: %.4f(us)" % (pdict['Tp'] * 1e6))
        print("---Range bandwidth: %.4f(MHz)" % (pdict['Br'] * 1e-6))
        print("---Range sampling rate: %.4f(MHz)" % (pdict['Fsr'] * 1e-6))
        print("---Range sampling time: %.4f(us)" % (pdict['Tr'] * 1e6))
        print("---Range over sampling factor: %.4f" % pdict['osfr'])
        print("---Range start sampling time: %.4f(us)" % (pdict['tstart'] * 1e6))
        print("---Range stop sampling time: %.4f(us)" % (pdict['tend'] * 1e6))
        print("---Scene area (m): [%.4f, %.4f, %.4f, %.4f]" % (pdict['SceneArea'][
              0], pdict['SceneArea'][1], pdict['SceneArea'][2], pdict['SceneArea'][3]))
        print("---Scene center (m): [%.4f, %.4f]" %
              (pdict['SceneCenter'][0], pdict['SceneCenter'][1]))
        print("---Beam area (m): [%.4f, %.4f, %.4f, %.4f]" % (pdict['BeamArea'][
            0], pdict['BeamArea'][1], pdict['BeamArea'][2], pdict['BeamArea'][3]))
        print("---Beam center (m): [%.4f, %.4f]" %
              (pdict['BeamCenter'][0], pdict['BeamCenter'][1]))
        print("---Echo size: Azimuth×Range=%d × %d" % (pdict['Na'], pdict['Nr']))
        print("---Azimuth resolution: %.4f(m)" % (pdict['AzimuthResolution']))
        print("---Slant range resolution: %.4f(m)" % (pdict['SlantRangeResolution']))
        print("---Ground range resolution: %.4f(m)" % (pdict['GroundRangeResolution']))
        print("---Image pixel resolution: %.4f(m) × %.4f(m)" %
              (pdict['ImagePixelResolution'][0], pdict['ImagePixelResolution'][1]))
        print("---Minimum slant range: %.4f(m)" % (pdict['Rnear']))
        print("---Maximum slant range: %.4f(m)" % (pdict['Rfar']))
        print("---Sampling window start time: %.4f(us)" % (pdict['SWST'] * 1e6))
        print("---Synthetic aperture size: %.4f(m)" % (pdict['Lsar']))
        print("---Synthetic aperture time: %.4f(s)" % (pdict['Tsar']))
        print("---Synthetic aperture pixels: %d" % (pdict['Nsar']))
        print("---Out compute_sar_parameters------------------------------------------------------")

    return pdict



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : multilook_process.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th


def multilook_spatial(Sslc, nlooks):
    r"""spatial multilook processing

    spatial averaging in azimuth or range direction.

    Args:
        Sslc (Tensor): Processed single look complex (or intensity) sar data tensor with size :math:`N_a×N_r`.
        nlooks (tuple or list): The number of looks in azimuth and range direction, [na, nr] or (na, nr).

    Returns:
        Smlc (Tensor): Processed multi-look complex tensor.

    """

    Na, Nr = Sslc.shape
    if nlooks is None:
        return Sslc

    # spatial averaging
    if nlooks[0] > 1:
        Smlca = th.zeros((int(Na / nlooks[0]), Nr), dtype=Sslc.dtype, device=Sslc.device)
        index = 0
        for a in range(0, Na - nlooks[0], nlooks[0]):
            Smlca[index, :] = th.mean(Sslc[a:a + (nlooks[0] - 1), :], axis=0)
            index += 1
    else:
        Smlca = Sslc

    Na, Nr = Smlca.shape

    if nlooks[1] > 1:
        Smlc = th.zeros((Na, int(Nr / nlooks[1])), dtype=Sslc.dtype, device=Sslc.device)
        index = 0
        for r in range(0, Nr - nlooks[1], nlooks[1]):
            Smlc[:, index] = th.mean(Smlca[:, r:r + (nlooks[1] - 1)], axis=1)
            index += 1
    else:
        Smlc = Smlca

    return Smlc


if __name__ == '__main__':

    Na, Nr = (1025, 256)
    real = th.randn(Na, Nr)
    imag = th.randn(Na, Nr)
    print(real.shape, imag.shape)
    Sslc = real + 1j * imag
    Smlc = multilook_spatial(Sslc, nlooks=(4, 1))

    print(Smlc.shape)



def readrcd(filepath, decfmtf, D, offset=0, endian='>'):
    """read file record

    read record information from a file. see :func:`read_ers_sar_raw` for reading ERS SAR raw data.

    Parameters
    ----------
    filepath : str
        file path string
    decfmtf : function
        format decoding function. see :func:`decfmtfers` for example.
    D : dict
        record descriptor dict, each value will be rewrited after reading.
    offset : int, optional
        record offset, read from the offset-th Byte (the default is 0)
    endian : str, optional
        endian, ``'<'`` --> little, ``'>'`` -- > big (the default is '<', which means little endian)

    Returns
    -------
    state : number
        reading status. 0 --> OK and done; 1 --> no such file; 2 --> bad record

    """

def readrcd1item(filepath, decfmtf, adrfmt, offset=0, endian='>'):
    """read one file record item

    read one record item information from a file. see :func:`read_ceos_sar_raw` for reading ERS SAR raw data.

    Parameters
    ----------
    filepath : str
        file path string
    decfmtf : function
        format decoding function. see :func:`decfmtfers` for example.
    adrfmt : tuple
        address and formation of one record item, a tuple with two
        elements: ``(adr, fmt)``, where, ``adr`` = (start address, end address) is a tuple and specifies
        the address (unit Bytes), ``fmt`` is the formation string.
    offset : int, optional
        record address offset, begin from the offset-th Byte (the default is 0)
    endian : str, optional
        endian, ``'<'`` --> little, ``'>'`` -- > big (the default is '<', which means little endian)

    Returns
    -------
    v : {}
        readed record item value.

    """

def printrcd(D):
    """print record

    print record in ascend address order.

    Parameters
    ----------
    D : dict
        descriptor dict
    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : chirp_signal.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th

from torchbox.utils.const import *
from torchbox.dsp.normalsignals import rect


def chirp_tran(t, Tp, K, Fc, a=1.):

    return a * rect(t / Tp) * th.exp(2j * PI * Fc * t + 1j * PI * K * t**2)


def chirp_recv(t, Tp, K, Fc, a=1., g=1., r=1e3):
    Sr = 0.
    for gi, ri in zip(g, r):
        tau = 2. * ri / C
        tt = t - tau  # do not use t -= tau, this will change t out this function
        Sr += gi * rect(tt / Tp) * th.exp(2j * PI * Fc * tt + 1j * PI * K * tt**2)
    return Sr


class Chirp(th.nn.Module):

    def __init__(self, Tp, K, Fc=0., a=1.):
        self.Tp = Tp
        self.K = K
        self.Fc = Fc
        self.a = a
        self.C = 299792458.
        self.PI = 3.141592653589793

    def tran(self, t):
        return self.a * rect(t / self.Tp) * th.exp(2j * self.PI * self.Fc * t + 1j * self.PI * self.K * t**2)

    def recv(self, t, g, r):

        Sr = 0.
        for gi, ri in zip(g, r):
            td = 2. * ri / self.C
            tt = t - td  # do not use t -= td, this will change t out this function
            Sr += self.a * gi * rect(tt / self.Tp) * th.exp(2j * self.PI * self.Fc * tt + 1j * self.PI * self.K * tt**2)

        return Sr


if __name__ == '__main__':

    import matplotlib.pyplot as plt
    from torch.fft import fft, fftshift

    # ===Generate tansmitted and recieved signals
    # ---Setting parameters
    R = [1.e3, 2.e3, 3.e3]
    G = [0.5, 1.0, 0.8]

    EPS = 2.2e-32
    K = 4.1e+11
    Tp = 37.0e-06
    Br = abs(K) * Tp

    alp = 1.24588  # 1.1-1.4
    Fsr = alp * Br
    Fc = 5.3e9
    Fc = 0.

    Tsr = 2.1 * Tp
    Nsr = int(Fsr * Tsr)
    t = th.linspace(-Tsr / 2., Tsr / 2, Nsr)
    f = th.linspace(-Fsr / 2., Fsr / 2, Nsr)

    # ---Transmitted signal
    St = chirp_tran(t, Tp, K, Fc, a=1.)

    # ---Recieved signal
    Sr = chirp_recv(t, Tp, K, Fc, a=1., g=G, r=R)

    chirp = Chirp(Tp=Tp, K=K, Fc=Fc, a=1.)

    St = chirp.tran(t)
    Sr = chirp.recv(t, g=G, r=R)

    plt.figure()
    plt.subplot(221)
    plt.plot(t * 1e6, th.real(St))
    plt.plot(t * 1e6, th.imag(St))
    plt.xlabel('Time/us')
    plt.legend(['real', 'imag'])
    plt.subplot(222)
    plt.plot(t * 1e6, th.angle(St))
    plt.xlabel('Time/us')
    plt.subplot(223)
    plt.plot(t * 1e6, th.real(Sr))
    plt.plot(t * 1e6, th.imag(Sr))
    plt.xlabel('Time/us')
    plt.legend(['real', 'imag'])
    plt.subplot(224)
    plt.plot(t * 1e6, th.angle(Sr))
    plt.xlabel('Time/us')
    plt.show()


    # ---Frequency domain
    Yt = fftshift(fft(fftshift(St, dim=0), dim=0), dim=0)
    Yr = fftshift(fft(fftshift(Sr, dim=0), dim=0), dim=0)

    # ---Plot signals
    plt.figure(figsize=(10, 8))
    plt.subplot(221)
    plt.plot(t * 1e6, th.real(St))
    plt.grid()
    plt.title('Real part')
    plt.xlabel('Time/μs')
    plt.ylabel('Amplitude')
    plt.subplot(222)
    plt.plot(t * 1e6, th.imag(St))
    plt.grid()
    plt.title('Imaginary part')
    plt.xlabel('Time/μs')
    plt.ylabel('Amplitude')
    plt.subplot(223)
    plt.plot(f, th.abs(Yt))
    plt.grid()
    plt.title('Spectrum')
    plt.xlabel('Frequency/Hz')
    plt.ylabel('Amplitude')
    plt.subplot(224)
    plt.plot(f, th.angle(Yt))
    plt.grid()
    plt.title('Spectrum')
    plt.xlabel('Frequency/Hz')
    plt.ylabel('Phase')
    plt.subplots_adjust(left=0.08, bottom=0.06, right=0.98, top=0.96, wspace=0.19, hspace=0.25)

    plt.show()

    plt.figure(figsize=(10, 8))
    plt.subplot(221)
    plt.plot(t * 1e6, th.real(Sr))
    plt.grid()
    plt.title('Real part')
    plt.xlabel('Time/μs')
    plt.ylabel('Amplitude')
    plt.subplot(222)
    plt.plot(t * 1e6, th.imag(Sr))
    plt.grid()
    plt.title('Imaginary part')
    plt.xlabel('Time/μs')
    plt.ylabel('Amplitude')
    plt.subplot(223)
    plt.plot(f, th.abs(Yr))
    plt.grid()
    plt.title('Spectrum')
    plt.xlabel('Frequency/Hz')
    plt.ylabel('Amplitude')
    plt.subplot(224)
    plt.plot(f, th.angle(Yr))
    plt.grid()
    plt.title('Spectrum')
    plt.xlabel('Frequency/Hz')
    plt.ylabel('Phase')
    plt.subplots_adjust(left=0.08, bottom=0.06, right=0.98, top=0.96, wspace=0.19, hspace=0.25)

    plt.show()



class SarData(object):
    r"""SAR data class.
        rawdata
        image
        store
        read

    """

    def name(self):
        ...

    def name(self, value):
        ...

    def rawdata(self):
        ...

    def rawdata(self, value):
        ...

    def image(self):
        ...

    def image(self, value):
        ...

    def store(self, sarplat, file):
        ...

    def read(self, file):
        r"""Read SAR data file

        Read SAR data file (``.pkl`` or ``.mat``)

        Parameters
        ----------
        file : str
            SAR data file path}

        Returns
        -------
        sardata : Instance SarData
            The SAR raw data.

        sarplat : Instance of SarPlat
            The SAR platform for obtain the SAR data.
        """





def _compensation(X, phi, Na, Nr, ftshift=False, isfft=True):
    ...

class AutoFocusFFO(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Nb=None, ftshift=False, trainable=True):
        ...

    def forward(self, X, isfft=True):
        ...

    def imaging(self, Xc):
        ...

    def get_param(self):
        ...

    def param_init(self, pa=None, pr=None):




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : sensor_airbone.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from torchbox.utils.const import *

SENSOR_AIRBONE = {

    'AIRSAR-P': {
        'Fc': 0.45e9,  # Hz
        'H': 8000,  # height in m
        'V': 215.0,  # velocity in m/s
        'Tp': 5e-6,  # Range pulse length in seconds
        'Kr': 8e+12,  # FM rate of radar pulse (Hz/s)
        'Lr': 12,  # antenna length (range) in m
        'La': 15.0,  # antenna length (azimuth) in m
        'PRF': 1800.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 48.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [2048, 2048],
        'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        # 'As': 0.05 * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 23.4 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Space3',
    },

    'UAVSAR': {
        'Fc': 1.27e9,  # Hz
        'H': 800000,  # height in m
        'V': 7153.0,  # velocity in m/s
        'Tp': 27e-6,  # Range pulse length in seconds
        'Kr': -28.e6 / 27.0e-6,  # FM rate of radar pulse (Hz/s)
        'Lr': 2.9,  # antenna length (range) in m
        'La': 8.9,  # antenna length (azimuth) in m
        'PRF': 1912.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 32.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        # 'EchoSize': [512, 1024],
        'EchoSize': [35345, 12040],
        'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        # 'As': 0.05 * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 34.3 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Space3',
    },

    'MSTAR': {
        'Fc': 9.60e9,  # Hz
        'H': 1480.75,  # height in m
        'V': 39.457031,  # velocity in m/s
        'Tp': 0.1e-6,  # Range pulse length in seconds
        'Kr': 591.0e6 / 0.1e-6,  # FM rate of radar pulse (Hz/s)
        'Lr': 0.6,  # antenna length (range) in m
        'La': 0.8,  # antenna length (azimuth) in m
        'PRF': 180.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 712.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        # 'EchoSize': [512, 1024],
        'EchoSize': [128, 128],
        'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        # 'As': 0.05 * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 17. * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'MSTAR',
    },

}



def rectangle(SceneArea, x0, y0, h, w, a=None, dx=None, dy=None, verbose=False):
    r"""rectangle area

    generates rectangle area

    Parameters
    ----------
    SceneArea : list or tuple
        Scene Area, unit m, [xmin,xmax,ymin,ymax]
    x0 : float
        x center of rectangle, unit m (x: horizontal, y: vertical)
    y0 : float
        y center of rectangle, unit m (x: horizontal, y: vertical)
    h : float
        height of rectangle, unit m (the default is None, which generates randomly)
    w : float
        width of rectangle, unit m (the default is None, which generates randomly)
    a : float, optional
        amplitude (the default is None, which generates randomly)
    dx : float, optional
        resolution in range (default: {1 / (xmax-xmin)})
    dy : float, optional
        resolution in azimuth (default: {1 / (ymax-ymin)})
    verbose : bool, optional
        show more log info (the default is False, which means does not show)

    Returns
    -------
    rects : tensor
        rectangle area
    """

def disc(SceneArea, x0, y0, r, a=None, dx=None, dy=None, verbose=False):
    r"""disc area

    generates disc area


    Parameters
    ----------
    SceneArea : list or tuple
        Scene Area, [xmin,xmax,ymin,ymax]
    x0 : int
        x center of disc (x: horizontal, y: vertical)
    y0 : int
        y center of disc (x: horizontal, y: vertical)
    r : int
        radius of disc
    a : int, optional
        amplitude of disc (default: {None})
    dx : float, optional
        resolution in range (default: {1 / (xmax-xmin)})
    dy : float, optional
        resolution in azimuth (default: {1 / (ymax-ymin)})
    verbose : bool, optional
        show more log info (the default is False, which means does not show)

    Returns
    -------
    ccas : tensor
        disk area
    """





def bdce_sf(Sr, Fsa, Fsr, rdflag=False, Nroff=0, isplot=False):
    r"""Baseband doppler centroid estimation by spectrum fitting

    Baseband doppler centroid estimation by spectrum fitting.

    Parameters
    ----------
    Sr : numpy array
        SAR signal :math:`N_a×N_r` in range-doppler domain (range frequency domain).
    Fsa : float
        Sampling rate in azimuth
    rdflag : bool
        Specifies whether the input SAR signal is in range-doppler domain. If not, :func:`dce_sf` excutes FFT in range direction.
    isplot : bool
        Whether to plot the estimated results.(Default: False)
    """

def bdce_api(Sr, Fsa, isplot=False):
    r"""Baseband doppler centroid estimation by average phase increment

    Baseband doppler centroid estimation by average phase increment.

    Parameters
    ----------
    Sr : numpy array
        SAR raw data or range compressed data
    Fsa : float
        Sampling rate in azimuth
    isplot : bool
        Whether to plot the estimated results.(Default: False)
    """

def bdce_madsen(Sr, Fsa, isplot=False):
    r"""Baseband doppler centroid estimation by madsen

    Baseband doppler centroid estimation bymadsen.

    Parameters
    ----------
    Sr : numpy array
        SAR raw data or range compressed data
    Fsa : float
        Sampling rate in azimuth
    isplot : bool
        Whether to plot the estimated results.(Default: False)
    """

def abdce_wda_ori(Sr, Fsa, Fsr, Fc, rate=0.9, isplot=False, islog=False):
    r"""Absolute and baseband doppler centroid estimation by wavelength diversity algorithm

    Absolute and baseband doppler centroid estimation by Wavelength Diversity Algorithm (WDA).

    <<合成孔径雷达成像_算法与实现>> p350.

    Parameters
    ----------
    Sr : numpy array
        SAR signal :math:`N_a×N_r` in range frequency domain.
    Fsa : float
        Sampling rate in azimuth.
    Fsr : float
        Sampling rate in range.
    """

def abdce_wda_opt(Sr, Fsr, Fsa, Fc, ncpb=None, tr=None, isfftr=False, isplot=False, islog=False):
    """Absolute and baseband doppler centroid estimation by wavelength diversity algorithm

    Absolute and baseband doppler centroid estimation by Wavelength Diversity Algorithm (WDA).

    <<合成孔径雷达成像_算法与实现>> p350.

    Parameters
    ----------
    Sr : 2d-tensor
        SAR signal :math:`N_a×N_r` in range frequency domain.
    Fsr : float
        Sampling rate in range, unit Hz.
    Fsa : float
        Sampling rate in azimuth, unit Hz.
    Fc : float
        Carrier frequency, unit Hz.
    ncpb : tuple or list, optional
        Number of cells per block, so we have blocks (int(Na/ncpb[0])) × (int(Nr/ncpb[1]))
        (the default is [Na, Nr], which means all).
    tr : 1d-tensor, optional
        Time in range (the default is None, which linspace(0, Nr, Nr)).
    isplot : bool, optional
        Whether to plot the estimation results (the default is False).
    isfftr : bool, optional
        Whether to do FFT in range (the default is False).

    Returns
    -------
    fadc : 2d-tensor
        Absolute doppler centroid frequency, which has the size specified by :attr:`ncpb`.
    fbdc : 2d-tensor
        Baseband doppler centroid frequency, which has the size specified by :attr:`ncpb`.
    Ma : 2d-tensor
        Doppler ambiguity number, which has the size specified by :attr:`ncpb`.
    """

def fullfadc(fdc, shape):




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : phase_error_model.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import math
import torch as th
import torchbox as tb
import torchsar as ts
from torch.nn.parameter import Parameter


class BarePhi(th.nn.Module):

    def __init__(self, Na, Nr, Nb=None, pa=None, pr=None, Ma=None, Mr=None, xa=None, xr=None, shift=False, trainable=True):
        super(BarePhi, self).__init__()

        if Nb is None:
            Nb = 1

        self.Na = Na
        self.Nr = Nr
        self.Nb = Nb
        self.Ma = Ma  # for remove linear trend
        self.Mr = Mr  # for remove linear trend
        self.shift = shift

        raise(ImportError('Not opened yet!'))


    def forward(self):
        returns = []
        raise(ImportError('Not opened yet!'))


    def get_param(self):
        param = []
        if self.Na is not None:
            param.append(self.pa)
        if self.Nr is not None:
            param.append(self.pr)
        return param

    def param_init(self, pa=None, pr=None):
        if self.pa is not None:
            if pa is None:
                pa = th.zeros(self.Nb, self.Na, dtype=self.pa.dtype, device=self.pa.device)
            self.pa.data = pa
        if self.pr is not None:
            if pr is None:
                pr = th.zeros(self.Nb, self.Nr, dtype=self.pr.dtype, device=self.pr.device)
            self.pr.data = pr


class PolyPhi(th.nn.Module):

    def __init__(self, Na, Nr, Nb=None, ca=None, cr=None, Ma=2, Mr=2, xa=None, xr=None, shift=False, trainable=True):
        super(PolyPhi, self).__init__()

        if type(Ma) is int:
            Ma = (2, Ma)
        if type(Mr) is int:
            Mr = (2, Mr)
        if Nb is None:
            Nb = 1

        self.Na = Na
        self.Nr = Nr
        self.Nb = Nb
        self.Ma = Ma
        self.Mr = Mr
        self.shift = shift

        raise(ImportError('Not opened yet!'))


    def forward(self):
        returns = []
        raise(ImportError('Not opened yet!'))


    def get_param(self):
        param = []
        if self.Na is not None:
            param.append(self.ca)
        if self.Nr is not None:
            param.append(self.cr)
        return param

    def param_init(self, ca=None, cr=None):
        if self.ca is not None:
            if ca is None:
                ca = th.zeros(self.Nb, self.Ma[1] - self.Ma[0] + 1, dtype=self.ca.dtype, device=self.ca.device)
            self.ca.data = ca
        if self.cr is not None:
            if cr is None:
                cr = th.zeros(self.Nb, self.Mr[1] - self.Mr[0] + 1, dtype=self.cr.dtype, device=self.cr.device)
            self.cr.data = cr


class DctPhi(th.nn.Module):

    def __init__(self, Na, Nr, Nb=None, ca=None, cr=None, Pa=4, Pr=4, shift=False, trainable=True):
        super(DctPhi, self).__init__()
        r"""

        .. math::
            \begin{array}{l}
            \phi_{e}(h)=\sum_{p=0}^{P} a(p) \cdot d c t(p) \cdot \cos \left[\frac{\pi\left(2
             h_{a}+1\right) p}{2 N}\right] \\
            a(p)=\left\{\begin{array}{l}
            1 / \sqrt{N} \quad p=0 \\
            \sqrt{2 / N} \quad p \neq 0
            \end{array},\right. \text { and }\left\{\begin{array}{l}
            p=0,1,2, \cdots \cdots \cdots, P \\
            h_{a}=[0,1,2, \cdots \cdots, N-1]
            \end{array}\right.
            \end{array}


        see Azouz, Ahmed Abd Elhalek, and Zhenfang Li. "Improved phase gradient autofocus algorithm based on segments of variable lengths and minimum-entropy phase correction." IET Radar, Sonar & Navigation 9.4 (2015): 467-479.


        Parameters
        ----------
        Na : int
            the number of samples in azimuth.
        Nr : int
            the number of samples in range.
        Nb : int, optional
            batch size (the default is None, which means 1)
        ca : tensor, optional
            DCT coefficients (the default is None, initialize with zeros)
        cr : tensor, optional
            DCT coefficients (the default is None, initialize with zeros)
        Pa : int, optional
            The number of DCT coefficients (the default is 2)
        Pr : int, optional
            The number of DCT coefficients (the default is 2)
        shift : bool, optional
            shift the frequency (the default is False, which means don't shift)
        trainable : bool, optional
            enable train (the default is True)
        """

        if Nb is None:
            Nb = 1

        self.Na = Na
        self.Nr = Nr
        self.Nb = Nb
        self.Pa = Pa
        self.Pr = Pr
        self.shift = shift

        if Na is not None:
            if ca is None:
                ca = th.zeros(Nb, Pa)
            self.ca = Parameter(ca, requires_grad=trainable)
            xa = th.linspace(0, Na, Na).reshape(1, Na)
            raise(ImportError('Not opened yet!'))


        if Nr is not None:
            if cr is None:
                cr = th.zeros(Nb, Pr)
            self.cr = Parameter(cr, requires_grad=trainable)
            xr = th.linspace(0, Nr, Nr).reshape(1, Nr)

            raise(ImportError('Not opened yet!'))


    def forward(self):
        returns = []
        raise(ImportError('Not opened yet!'))


    def get_param(self):
        param = []
        if self.Na is not None:
            param.append(self.ca)
        if self.Nr is not None:
            param.append(self.cr)
        return param

    def param_init(self, ca=None, cr=None):
        if self.ca is not None:
            if ca is None:
                ca = th.zeros(self.Nb, self.Pa, dtype=self.ca.dtype, device=self.ca.device)
            self.ca.data = ca
        if self.cr is not None:
            if cr is None:
                cr = th.zeros(self.Nb, self.Pr, dtype=self.cr.dtype, device=self.cr.device)
            self.cr.data = cr


class SinPhi(th.nn.Module):

    def __init__(self, Na, Nr, Nb=None, aa=None, fa=None, ar=None, fr=None, shift=False, trainable=True):
        super(SinPhi, self).__init__()
        r"""

        The sinusoidal phase error can be expressed as

        .. math::
            \Phi_{e}=\phi_{0} \sin \left(2 \pi f_{e} t\right)

        The Bessel approximation to a sinusoidal phase error is

        .. math::
            e^{j \phi_{0} \sin \left(2 \pi f_{e} t\right)}=\sum_{n=-\infty}^{\infty} J_{n}\left(\phi_{0}\right) e^{j 2 \pi n f_{e} t}

        For small :math:`\phi_{0}`, :math:`J_0 \simeq 1`, :math:`J_{1}\left(\phi_{0}\right)=-J_{-1}\left(\phi_{0}\right) \simeq \phi_{0} / 2`, it can be rewritten as

        .. math::
            e^{j \phi_{0} \sin \left(2 \pi f_{e} t\right)} & \simeq J_{0}\left(\phi_{0}\right)+
            J_{1}\left(\phi_{0}\right) e^{j 2 \pi f_{e} t}+J_{-1}\left(\phi_{0}\right) e^{-j 2 \
            pi f_{e} t} \\
            & \simeq 1+\frac{\phi_{0}}{2}\left(e^{j 2 \pi f_{e} t}-e^{-j 2 \pi f_{e} t}\right)


        [1] Satyaprasad, Shruthi B., et al. "Autofocusing SAR images using multistage wiener filter."
        2016 IEEE International Conference on Recent Trends in Electronics, Information & Communication
        Technology (RTEICT). IEEE, 2016.

        [2] Kim, Jin-Woo, et al. "Fast Fourier-Domain Optimization Using Hybrid $L_1 / L_{p}$-Norm for
        Autofocus in Airborne SAR Imaging." IEEE Transactions on Geoscience and Remote Sensing 57.10 (2019): 7934-7954.


        Parameters
        ----------
        Na : int
            the number of samples in azimuth.
        Nr : int
            the number of samples in range.
        Nb : int, optional
            batch size (the default is None, which means 1)
        aa : float, optional
            amplitude (the default is None, initialize with zeros)
        fa : float, optional
            frequency (the default is None, initialize with zeros)
        ar : float, optional
            amplitude (the default is None, initialize with zeros)
        fr : float, optional
            frequency (the default is None, initialize with zeros)
        shift : bool, optional
            shift the frequency (the default is False, which means don't shift)
        trainable : bool, optional
            enable train (the default is True)
        """

        if Nb is None:
            Nb = 1

        self.Na = Na
        self.Nr = Nr
        self.Nb = Nb
        self.shift = shift

        raise(ImportError('Not opened yet!'))


    def forward(self):
        returns = []
        raise(ImportError('Not opened yet!'))


    def get_param(self):
        param = []
        if self.Na is not None:
            param.append(self.aa)
            param.append(self.fa)
        if self.Nr is not None:
            param.append(self.ar)
            param.append(self.fr)
        return param

    def param_init(self, fa=None, aa=None, fr=None, ar=None):
        if self.aa is not None:
            if aa is None:
                aa = th.ones(self.Nb, 1, dtype=self.aa.dtype, device=self.aa.device)
            self.aa.data = aa
        if self.fa is not None:
            if fa is None:
                fa = th.zeros(self.Nb, 1, dtype=self.fa.dtype, device=self.fa.device)
            self.fa.data = fa

        if self.ar is not None:
            if ar is None:
                ar = th.ones(self.Nb, 1, dtype=self.ar.dtype, device=self.ar.device)
            self.ar.data = ar
        if self.fr is not None:
            if fr is None:
                fr = th.zeros(self.Nb, 1, dtype=self.fr.dtype, device=self.fr.device)
            self.fr.data = fr



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : make_targets.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
from torchbox.utils.const import *
from torchbox.dsp import normalsignals as sig
from torchbox import setseed


def gpoints(SceneArea, npoints, amin=0., amax=1., seed=None):
    """Generates point targets

    Parameters
    ----------
    SceneArea : tuple or list
        The area of scene.
    npoints : int
        The number of points.
    amin : float, optional
        The minimum amplitude. (default is 0)
    amax : float, optional
        The maximum amplitude. (default is 1)
    seed : int or None, optional
        The seed for random generator.

    Returns
    -------
    tensor
        A tensor contains coordinate and amplitude information.
    """
    if seed is not None:
        setseed(seed)
    xmin, xmax, ymin, ymax = SceneArea

    xs = th.rand(npoints, 1) * (xmax - xmin) + xmin
    ys = th.rand(npoints, 1) * (ymax - ymin) + ymin
    amps = th.rand(npoints, 1) * (amax - amin) + amin

    targets = th.zeros(npoints, 7)

    targets[:, [0]] = xs
    targets[:, [1]] = ys
    targets[:, [-1]] = amps

    return targets



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : doppler_rate_estimation.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import division, print_function
import torch as th
from torchbox.utils.const import *


def dre_geo(Wl, Vr, R, Ar=0.):
    """doppler rate estimation based on geometry

    doppler rate estimation based on geometry

    Parameters
    ----------
    Wl : float
        Wave length.
    Vr : float, list or array
        Equivalent velocity of radar.
    R : float, list or array
        Minimum slant range from radar to target.
    Ar : float, list or array
        Equivalent squint angle.
    """

    raise(ImportError('Not opened yet!'))





def extract_targets(x, region=None, thresh=None, cdim=None, isshow=False):
    r"""Extracts the pixels of targetss

    Parameters
    ----------
    x : tensor or numpy array
        The input image, if it's complex-valued, it's amplitude is used.
    region : list or None, optional
        The region ([top-left, bottom-right]) that contains targets for computing TBR.
        If :obj:`None`, :attr:`region` equals to ``([0, 0], [H, W])``, where :math:`H,W` 
        are the height and width of the image.
    thresh : float or None, optional
        The threshold targets.
    cdim : int or None, optional
        Specifies the complex axis of :attr:`x`. ``None`` --> complex or real
    isshow : bool, optional
        Show pixels of the extracted targets?
    """

def tbr(x, targets, region=None, cdim=None, isshow=False):
    r"""Target-to-Background Ratio (TBR)

    .. math::
        \begin{align}
        {\rm TBR} = 20{\rm log}_{10}\left(\frac{{\rm max}_{i\in{\mathbb T}}(|{\bf X}_i|)}{{(1/N_{\mathbb B})}\Sigma_{j\in \mathbb B}|{\bf X}_j|)}\right)
        \label{equ:TBR}
        \end{align}

    Parameters
    ----------
    x : tensor
        The input image, if it's complex-valued, it's amplitude is used.
    targets : list or tuple
        The targets pixels. ([row1, row2, ...], [col1, col2, ...])
    region : list, tuple or None, optional
        The region ([top-left, bottom-right]) that contains targets for computing TBR.
        If :obj:`None`, :attr:`region` equals to ``([0, 0], [H, W])``, where :math:`H,W` 
        are the height and width of the image.
    cdim : int or None, optional
        Specifies the complex axis of :attr:`x`. ``None`` --> complex or real
    isshow : bool, optional
        Show target mask? (default: False)
    """

def tbr2(X, tgrs, subrs=None, isshow=False):
    r"""Target-to-Background Ratio (TBR)

    .. math::
        \begin{align}
        {\rm TBR} = 20{\rm log}_{10}\left(\frac{{\rm max}_{i\in{\mathbb T}}(|{\bf X}_i|)}{{(1/N_{\mathbb B})}\Sigma_{j\in \mathbb B}|{\bf X}_j|)}\right)
        \label{equ:TBR}
        \end{align}

    Parameters
    ----------
    X : tensor
        The input image, if it's complex-valued, it's amplitude is used.
    tgrs : list, optional
        target regions:[[TG1], [TG2], ..., [TGn]], [TGk] = [lefttop, rightbottom]]
    subrs : list, optional
        sub regions:[[SUB1], [SUB2], ..., [SUBn]], [SUBk] = [lefttop, rightbottom]]
    isshow : bool, optional
        show target mask given by :attr:`tgrs` (default: False)

    Returns
    -------
    TBR : float
        Target-to-Background Ratio (TBR)

    Raises
    ------
    TypeError
        tgrs mast be given
    """





class RangeCompress(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Tp, Fsr, Kr, Fc, trainable=True, dtype=th.float32):
        ...

    def forward(self, X):
        ...

class AzimuthCompress(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Tp, Fsa, Ka, Fc, trainable=True, dtype=th.float32):
        ...

    def forward(self, X):
        ...

class AzimuthCompressLinearFit(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Tp, Fsa, Ka, Fc, trainable=True, dtype=th.float32):
        ...

    def forward(self, X):




def gpoints(SceneArea, npoints, amin=0., amax=1., seed=None):
    """Generates point targets

    Parameters
    ----------
    SceneArea : tuple or list
        The area of scene.
    npoints : int
        The number of points.
    amin : float, optional
        The minimum amplitude. (default is 0)
    amax : float, optional
        The maximum amplitude. (default is 1)
    seed : int or None, optional
        The seed for random generator.

    Returns
    -------
    tensor
        A tensor contains coordinate and amplitude information.
    """





def vga_gain_compensation(S, V, mod='linear', fact=1.0):
    r"""vga gain compensation

    vga gain compensation

    .. math::
       \begin{aligned}
       {\bm F} &= (λ 10^{{\bm V}/20})\\
       {\bm S}_{c} &= {\bm F} \odot {\bm S}
       \end{aligned}

    Parameters
    ----------
    S : torch tensor
        :attr:`S` is an :math:`N_a×N_r×2` array, where, :math:`S[:,:,0]` is the I signal
        and :math:`S[:,:,1]` is the Q signal.
    V : torch tensor
        :attr:`S` is an :math:`N_a×N_r` or :math:`N_a×1` VGA gain array, the gain values are in
        ``dB`` unit.
    mod : str, optional
        compensation mode (the default is 'linear')
    fact : number, optional
        fact is the factor :math:`\lambda` (the default is 1.0)

    Returns
    -------
    torch tensor
        compensated signal, :math:`N_a×N_r×2` array.
    """

# def vga_gain_compensation(S, V, mod='linear', fact=1.0):
#     r"""vga gain compensation

#     vga gain compensation

#     .. math::
#        \begin{aligned}
#        {\bm F} &= (λ 10^{{\bm V}/20})\\
#        {\bm S}_{c} &= {\bm F} \odot {\bm S}
#        \end{aligned}

#     Parameters
#     ----------
#     S : array
#         :attr:`S` is an :math:`N_a×N_r×2` array, where, :math:`S[:,:,0]` is the I signal
#         and :math:`S[:,:,1]` is the Q signal.
#     V : array
#         :attr:`S` is an :math:`N_a×N_r` or :math:`N_a×1` VGA gain array, the gain values are in
#         ``dB`` unit.
#     mod : str, optional
#         compensation mode (the default is 'linear')
#     fact : int, optional
#         fact is the factor :math:`\lambda` (the default is 1.0)

#     Returns
#     -------
#     array
#         compensated signal, :math:`N_a×N_r×2` array.
#     """





class BarePhi(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Nb=None, pa=None, pr=None, Ma=None, Mr=None, xa=None, xr=None, shift=False, trainable=True):
        ...

    def forward(self):
        ...

    def get_param(self):
        ...

    def param_init(self, pa=None, pr=None):
        ...

class PolyPhi(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Nb=None, ca=None, cr=None, Ma=2, Mr=2, xa=None, xr=None, shift=False, trainable=True):
        ...

    def forward(self):
        ...

    def get_param(self):
        ...

    def param_init(self, ca=None, cr=None):
        ...

class DctPhi(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Nb=None, ca=None, cr=None, Pa=4, Pr=4, shift=False, trainable=True):
        r"""

        .. math::
            \begin{array}{l}
            \phi_{e}(h)=\sum_{p=0}^{P} a(p) \cdot d c t(p) \cdot \cos \left[\frac{\pi\left(2
             h_{a}+1\right) p}{2 N}\right] \\
            a(p)=\left\{\begin{array}{l}
            1 / \sqrt{N} \quad p=0 \\
            \sqrt{2 / N} \quad p \neq 0
            \end{array},\right. \text { and }\left\{\begin{array}{l}
            p=0,1,2, \cdots \cdots \cdots, P \\
            h_{a}=[0,1,2, \cdots \cdots, N-1]
            \end{array}\right.
            \end{array}


        see Azouz, Ahmed Abd Elhalek, and Zhenfang Li. "Improved phase gradient autofocus algorithm based on segments of variable lengths and minimum-entropy phase correction." IET Radar, Sonar & Navigation 9.4 (2015): 467-479.


        Parameters
        ----------
        Na : int
            the number of samples in azimuth.
        Nr : int
            the number of samples in range.
        Nb : int, optional
            batch size (the default is None, which means 1)
        ca : tensor, optional
            DCT coefficients (the default is None, initialize with zeros)
        cr : tensor, optional
            DCT coefficients (the default is None, initialize with zeros)
        Pa : int, optional
            The number of DCT coefficients (the default is 2)
        Pr : int, optional
            The number of DCT coefficients (the default is 2)
        shift : bool, optional
            shift the frequency (the default is False, which means don't shift)
        trainable : bool, optional
            enable train (the default is True)
        """

    def forward(self):
        ...

    def get_param(self):
        ...

    def param_init(self, ca=None, cr=None):
        ...

class SinPhi(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Nb=None, aa=None, fa=None, ar=None, fr=None, shift=False, trainable=True):
        r"""

        The sinusoidal phase error can be expressed as

        .. math::
            \Phi_{e}=\phi_{0} \sin \left(2 \pi f_{e} t\right)

        The Bessel approximation to a sinusoidal phase error is

        .. math::
            e^{j \phi_{0} \sin \left(2 \pi f_{e} t\right)}=\sum_{n=-\infty}^{\infty} J_{n}\left(\phi_{0}\right) e^{j 2 \pi n f_{e} t}

        For small :math:`\phi_{0}`, :math:`J_0 \simeq 1`, :math:`J_{1}\left(\phi_{0}\right)=-J_{-1}\left(\phi_{0}\right) \simeq \phi_{0} / 2`, it can be rewritten as

        .. math::
            e^{j \phi_{0} \sin \left(2 \pi f_{e} t\right)} & \simeq J_{0}\left(\phi_{0}\right)+
            J_{1}\left(\phi_{0}\right) e^{j 2 \pi f_{e} t}+J_{-1}\left(\phi_{0}\right) e^{-j 2 \
            pi f_{e} t} \\
            & \simeq 1+\frac{\phi_{0}}{2}\left(e^{j 2 \pi f_{e} t}-e^{-j 2 \pi f_{e} t}\right)


        [1] Satyaprasad, Shruthi B., et al. "Autofocusing SAR images using multistage wiener filter."
        2016 IEEE International Conference on Recent Trends in Electronics, Information & Communication
        Technology (RTEICT). IEEE, 2016.

        [2] Kim, Jin-Woo, et al. "Fast Fourier-Domain Optimization Using Hybrid $L_1 / L_{p}$-Norm for
        Autofocus in Airborne SAR Imaging." IEEE Transactions on Geoscience and Remote Sensing 57.10 (2019): 7934-7954.


        Parameters
        ----------
        Na : int
            the number of samples in azimuth.
        Nr : int
            the number of samples in range.
        Nb : int, optional
            batch size (the default is None, which means 1)
        aa : float, optional
            amplitude (the default is None, initialize with zeros)
        fa : float, optional
            frequency (the default is None, initialize with zeros)
        ar : float, optional
            amplitude (the default is None, initialize with zeros)
        fr : float, optional
            frequency (the default is None, initialize with zeros)
        shift : bool, optional
            shift the frequency (the default is False, which means don't shift)
        trainable : bool, optional
            enable train (the default is True)
        """

    def forward(self):
        ...

    def get_param(self):
        ...

    def param_init(self, fa=None, aa=None, fr=None, ar=None):




def dre_geo(Wl, Vr, R, Ar=0.):
    """doppler rate estimation based on geometry

    doppler rate estimation based on geometry

    Parameters
    ----------
    Wl : float
        Wave length.
    Vr : float, list or array
        Equivalent velocity of radar.
    R : float, list or array
        Minimum slant range from radar to target.
    Ar : float, list or array
        Equivalent squint angle.
    """





def tgs2sar_td(targets, pdict, mode='Target', device='cpu'):
    """Time-domain simulation

    SAR raw data simulation by time-domain method.

    Parameters
    ----------
    targets : tensor
        A 2d-tensor contains the information of targets,
        each row is a target [x, y, z, vx, vy, vz, a].
    pdict : dict
        The SAR platform parameters.
    mode : str, optional
        Simulation mode.
    device : str, optional
        Specifies which device to be used.

    Returns
    -------
    tensor
        Simulated SAR raw data tensor.
    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : channel_process.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import math
import torch as th


def iq_correct(Sr):
    r"""IQ Correction performs the I Q data correction

    IQ Correction performs the I Q data correction

    - I/Q bias removal
    - I/Q gain imbalance correction
    - I/Q non-orthogonality correction

    see "Sentinel-1-Level-1-Detailed-Algorithm-Definition".

    Args:
        Sr (Tensor): SAR raw data matrix :math:`{\bm S}_r \in {\mathbb R}^{N_a×N_r×2}`

    Returns:
        Sr (Tensor): Corrected SAR raw data.
        Flag : dict

    """

    Na, Nr, _ = Sr.shape

    raise(ImportError('Not opened yet!'))



if __name__ == '__main__':

    Na, Nr = (4, 5)
    theta = th.rand(Na, Nr)

    Is = th.cos(theta)
    Qs = th.sin(theta)
    Sr = th.zeros((Na, Nr, 2))
    Sr[:, :, 0] = Is
    Sr[:, :, 1] = Qs

    print(Sr)
    Sr, Flag = iq_correct(Sr)
    print(Sr)
    print(Flag)



def multilook_spatial(Sslc, nlooks):
    r"""spatial multilook processing

    spatial averaging in azimuth or range direction.

    Args:
        Sslc (Tensor): Processed single look complex (or intensity) sar data tensor with size :math:`N_a×N_r`.
        nlooks (tuple or list): The number of looks in azimuth and range direction, [na, nr] or (na, nr).

    Returns:
        Smlc (Tensor): Processed multi-look complex tensor.

    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : sarplat.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import logging
import torch as th
import numpy as np
from torchbox.utils.const import *
from torchsar.sharing.beamwidth_footprint import azimuth_beamwidth, compute_range_beamwidth2, cr_footprint
from torchsar.sharing.slant_ground_range import min_slant_range

AirboneSatelliteHeightBoundary = 60e3  # 60 km


class SarPlat(object):
    r"""SAR platform class.
        sensor
        acquisition
        params
    """

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        logging.info("---Check and set SAR platform name...")

        if not isinstance(value, str):
            raise ValueError('name must be a string!')
        self._name = value
        logging.info("---Done!")

    @property
    def sensor(self):
        return self._sensor

    @sensor.setter
    def sensor(self, value):
        logging.info("---Check and set SAR sensor parameters...")

        if not isinstance(value, dict):
            raise ValueError('sensor must be a dict!')
        if hasattr(self, '_sensor'):
            for k, v in value.items():
                self._sensor[k] = v
        else:
            self._sensor = value

        self._sensor['Br'] = self._sensor['Tp'] * np.abs(self._sensor['Kr'])  # chirp rate in Hz/s
        self._sensor['Wl'] = C / self._sensor['Fc']  # wave length in m

        Re = Ree
        Rs = self._sensor['H'] + Re
        # Rs = self._sensor['H'] + Ree
        self._sensor['Rs'] = Rs
        if value['H'] < AirboneSatelliteHeightBoundary:
            logging.info("---Airbone.")
            self._sensor['Vs'] = self._sensor['V']
            self._sensor['Vr'] = self._sensor['V']
            self._sensor['Vg'] = self._sensor['V']
            self._sensor['To'] = -1
            self._sensor['Ws'] = -1
        else:
            logging.info("---Satellite.")
            # Earth Bending Geometry (Vg < Vr < Vs, Vr = \sqrt{Vg*Vs})
            self._sensor['To'] = 2 * PI * np.sqrt((Rs)**3 / Ge)
            self._sensor['Ws'] = 2 * PI / self._sensor['To']
            self._sensor['Vs'] = Rs * self._sensor['Ws']
            self._sensor['Vr'] = self._sensor['V']
            # Equivalent velocity (Linear geometry)
            self._sensor['Vg'] = self._sensor['Vr']**2 / self._sensor['Vs']  # ground velocity
        Vr = self._sensor['Vs'] * np.sqrt(Re / Rs)
        logging.info("---Computed equivalent velocity is " + str(Vr) +
                     ", setted equivalent velocity is " + str(self._sensor['Vr']) + "(used).")
        self._sensor['Vs'] = Vr / np.sqrt(Re / Rs)  # reset
        logging.info("---Done!")

    @property
    def acquisition(self):
        return self._acquisition

    @acquisition.setter
    def acquisition(self, value):
        logging.info("---Check and set acquisition parameters...")

        if value is not None:
            if not isinstance(value, dict):
                raise ValueError('acquisition must be a dict!')
            if hasattr(self, '_acquisition'):
                for k, v in value.items():
                    self._acquisition[k] = v
            else:
                self._acquisition = value

        H = self._sensor['H']
        V = self._sensor['V']
        Vr = self._sensor['Vr']
        Vs = self._sensor['Vs']
        PRF = self._sensor['PRF']
        Fsr = self._sensor['Fs']
        Tp = self._sensor['Tp']
        Rs = self._sensor['Rs']

        As = value['As']
        Ad = value['Ad']
        Aon = value['Aon']
        Aba = value['Aba']
        Abr = value['Abr']
        SA = value['SceneArea']
        PC = value['PlatCenter']
        ES = value['EchoSize']
        SV = value['StateVector']
        if SV is None:
            SV = []

        if Aba is None:
            Aba = azimuth_beamwidth(self._sensor['Wl'], self._sensor['La'])

        # As: squint angle in earth curve geometry, Ar: squint angle in line geometry
        Ar = (Vs / Vr) * As

        if PC is None:
            PC = [0.0, 0.0, self._sensor['H']]

        if value['EchoSize'] is None:
            logging.error("~~~EchoSize should be specified!")

        SAV = [0, 0, 0, 0]
        BAV = [0, 0, 0, 0]
        if SA is None:
            SA = [None, None, None, None]

        if Abr is None:
            Nr = ES[1]
            Abr = compute_range_beamwidth2(Nr, Fsr, H, Aon)
        if Ad is None:
            Ad = Aon + Abr / 2.
            # pdict['Abr'] = math.acos(pdict['H'] / pdict['Rfar']) - abs(pdict['Aon'])
            Ad = np.sign(Aon) * (PI / 2. - (abs(Aon) + Abr / 2.))

        if (Ad - Abr / 2.0) <= 0 or (Ad + Abr / 2.0) >= np.pi / 2.0:
            logging.error(r"~~~The antenna elevation beamwidth Abr is too large!")

        Rnear = H / (np.cos(Aon) + EPS)
        Rfar = H / (np.cos(Aon + np.sign(Aon) * Abr) + EPS)
        FPr = np.sqrt(Rfar**2 + Rnear**2 - 2. * Rnear * Rfar * np.cos(Abr) + EPS)

        Rsc = 0.5 * np.sqrt(2. * Rfar**2 + 2. * Rnear**2 - FPr**2 + EPS)
        Rs0 = Rsc * np.cos(Ar)

        Ysc = Rsc * np.sin(Ar)
        Xsc = np.sqrt(np.abs(Rs0**2 - H**2))
        SC = [Xsc, Ysc, 0.0]

        Rbc = H / (np.sin(Ad) + EPS)
        Rb0 = Rbc * np.cos(Ar)

        Ybc = Rbc * np.sin(Ar)
        Xbc = np.sqrt(np.abs(Rb0**2 - H**2))
        BC = [Xbc, Ybc, 0.0]

        if (SA[0] is None) or (SA[1] is None):
            Ynear = Rnear * np.sin(Ar)
            Yfar = Rfar * np.sin(Ar)
            Xnear = np.sqrt((Rnear ** 2 - H**2) - Ynear ** 2 + EPS)
            Xfar = np.sqrt((Rfar ** 2 - H**2) - Yfar ** 2 + EPS)
            SAV[0] = Xnear - Xsc
            SAV[1] = Xfar - Xsc
        else:
            SAV[0] = SA[0]
            SAV[1] = SA[1]

        if (SA[2] is None) or (SA[3] is None):
            Na, Nr = ES
            Ta = Na / PRF
            SAV[2] = -Ta * V / 2.0
            SAV[3] = Ta * V / 2.0
        else:
            SAV[2] = SA[2]
            SAV[3] = SA[3]

        Na, Nr = ES
        Ta = Na / PRF
        FPr0 = H / np.tan(Ad + Abr / 2.) - H / np.tan(Ad)
        FPr1 = H / np.tan(Ad - Abr / 2.) - H / np.tan(Ad)
        BAV = (FPr0, FPr1, -Ta * V / 2.0, Ta * V / 2.0)

        if H < AirboneSatelliteHeightBoundary:
            logging.info("---Airbone.")
            Ai = PI / 2. - Ad
        else:
            Ao = np.arccos((Rs**2 + Rea**2 - Rbc**2) / (2. * Rs * Rea))
            Ai = (PI / 2. - Ad) + Ao
            logging.info("---Satellite.")
        logging.info('---Done!')

        value['StateVector'] = SV
        value['PlatCenter'] = np.array(PC)
        value['SceneCenter'] = np.array(SC)
        value['BeamCenter'] = np.array(BC)
        value['SceneArea'] = np.array(SAV)
        value['BeamArea'] = np.array(BAV)
        value['Rbc'] = Rbc
        value['Rb0'] = Rb0
        value['Rsc'] = Rsc
        value['Rs0'] = Rs0
        value['Rnear'] = Rnear
        value['Rfar'] = Rfar
        value['Aba'] = Aba
        value['Abr'] = Abr
        value['Ar'] = Ar
        value['Ai'] = Ai
        value['Ad'] = Ad
        value['Aon'] = Aon

        self._acquisition = value

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, value):

        logging.info("---Check and set algorithm parameters...")

        if value is not None:
            if not isinstance(value, dict):
                raise ValueError('params must be a dict!')
            if hasattr(self, '_params'):
                for k, v in value.items():
                    self._params[k] = v
            else:
                self._params = value

            return self._params

        GM = self._params['GeometryMode']
        if not isinstance(self._sensor, dict):
            raise ValueError('sensor must be a dict!')
        if not isinstance(self._acquisition, dict):
            raise ValueError('acquisition must be a dict!')

        H = self._sensor['H']
        V = self._sensor['V']
        Vr = self._sensor['Vr']
        Vg = self._sensor['Vg']
        Vs = self._sensor['Vs']
        La = self._sensor['La']
        Fsa = self._sensor['PRF']  # sampling rate in Azimuth dimension
        Fsr = self._sensor['Fs']  # sampling rate in range dimension
        Tp = self._sensor['Tp']  # Pulse repetition period
        Fc = self._sensor['Fc']
        Wl = self._sensor['Wl']
        Kr = self._sensor['Kr']
        La = self._sensor['La']

        Rbc = self._acquisition['Rbc']
        Rb0 = self._acquisition['Rb0']
        Rsc = self._acquisition['Rsc']
        Rs0 = self._acquisition['Rs0']
        Rbc = self._acquisition['Rbc']
        Rsc = self._acquisition['Rsc']
        Rnear = self._acquisition['Rnear']
        Rfar = self._acquisition['Rfar']

        PC = self._acquisition['PlatCenter']
        SC = self._acquisition['SceneCenter']
        BC = self._acquisition['BeamCenter']
        SA = self._acquisition['SceneArea']
        BA = self._acquisition['BeamArea']
        ES = self._acquisition['EchoSize']
        As = self._acquisition['As']
        Ar = self._acquisition['Ar']
        Ad = self._acquisition['Ad']
        Aon = self._acquisition['Aon']
        Aba = self._acquisition['Aba']
        Abr = self._acquisition['Abr']
        Ai = self._acquisition['Ai']

        if GM == 'BG':
            Xc = BC[0]
            Yc = BC[1]
            t0 = 2 * Rbc / C
            R0 = Rb0
        if GM == 'SG':
            t0 = 2 * Rsc / C
            Xc = SC[0]
            Yc = SC[1]
            R0 = Rs0

        Lsar = self._sensor['Wl'] * Rnear / La  # Rb0, Rbc, R?
        # Ka = (2.0 * Vr**2) / (Wl * Rs0)
        Ka = (-2.0 * Vr**2 * np.cos(Ar)**3) / (Wl * Rnear)
        Tsar = Lsar / Vr
        Ba = np.abs(Ka * Tsar)

        if self._sensor['PRF'] is None:
            PRF = (1.25 * Ba)
        else:
            PRF = self._sensor['PRF']

        Fsa = PRF
        if Fsa / Ba < 1.1 or Fsa / Ba > 2.0:
            print(Fsa, Ba)
            logging.warning(
                "~~~Sampling rate should be in range [1.1, 2.0]*Ba, \r\n   Fsa %f Ba %f!" % (Fsa, Ba))

        PRT = 1.0 / PRF  # azimuth sampling
        Tsa = 1.0 / PRF

        if self._sensor['Fs'] is None:
            Fs = (1.905 * self._sensor['Br'])
        else:
            Fs = self._sensor['Fs']

        Fsr = Fs
        Tsr = 1.0 / Fsr

        Na, Nr = ES

        Ta = Na * 1.0 / Fsa
        ta = np.linspace(-Ta / 2.0, Ta / 2.0, Na)
        fa = np.linspace(-Fsa / 2.0, Fsa / 2.0, Na)

        tnear = Rnear / C
        tfar = Rfar / C
        tstart = 2.0 * tnear
        # tend = 2.0 * tfar + Tp
        tend = 2.0 * tfar

        Tr = tend - tstart
        if abs(Nr - Tr * Fsr) > 1:
            Abr = compute_range_beamwidth2(Nr, Fsr, H, Aon)
            logging.error("~~~Nr should be Tr×Fsr= % f×% f= % f\r\n\
                          or Fsr should be Nr÷Tr= % f\\% f= % f\r\n\
                          or Abr should be Abr= % f degree!" %
                          (Tr, Fsr, Tr * Fsr, Nr, Tr, Nr / Tr, Abr * 180. / PI))
            raise ValueError("Nr ≠ Tr × Fsr! %f ≠ %f × %f = %f" %
                             (Nr, Tr, Fsr, Tr * Fsr))

        tr = np.linspace(tstart, tend, Nr)
        # tr = np.linspace(-Tr / 2.0, Tr / 2.0, Nr) + t0
        fr = np.linspace(-Fsr / 2.0, Fsr / 2.0, Nr)

        # tac = -Yc / Vg
        # tac = -R0 * np.tan(As) / Vg
        tac = -R0 * np.tan(Ar) / Vr

        # both are ok, equal
        fadc = 2.0 * Vr * np.sin(Ar) / Wl
        # fadc = 2.0 * Vs * np.sin(As) / Wl

        Bdop = 0.886 * (2 * Vs * np.cos(As) / La)

        xmin, xmax, ymin, ymax = SA + [Xc, Xc, Yc, Yc]

        if 1.0 / Fsa - Tp < Tr:
            logging.info(
                "r~~~Sampling rate should be lower to avoid range ambiguity \r\n   \
                (1.0 ÷ Fsa - Tp = 1.0 ÷ %f - %f > Tr = %f" % (Fsa, Tp, Tr))

        self._sensor['PRF'] = PRF
        self._sensor['Fs'] = Fs

        FPa = cr_footprint(Wl, H, La, Ad)
        # FPr = ar_footprint(Wl, H, Lr, Ad)
        FPr = SA[1] - SA[0]
        BWa = azimuth_beamwidth(Wl, La)

        PRFmin = 2.0 * V / H
        PRFmax = C / (2.0 * FPa * np.sin(Ad))

        if Lsar > FPa:
            logging.warning(
                r"~~~The synthetic aperture length Lsar=%s is large than \r\n   \
                the cross range footprint size Rcr=%s which maks a point outside the beam footprint." % (Lsar, FPa))
        if PRF < PRFmin or PRF > PRFmax:
            logging.warning(r"~~~To avoid grating lobe issues, PRF=%s should in [%s, %s]."
                            % (PRF, PRFmin, PRFmax))
        param = dict()
        param['GeometryMode'] = GM  # geometry mode: beam/scene
        param['Rb0'] = Rb0  # beam center distance
        param['Rbc'] = Rbc  # beam center distance
        param['Tp'] = Tp  # Pulse repetition period
        param['PRT'] = PRT  # azimuth
        param['PRF'] = PRF  # azimuth
        param['Fsa'] = Fsa  # azimuth, Fsa = PRF
        param['Fs'] = Fs  # range
        param['Fsr'] = Fsr  # range, Fsr = Fs
        param['Na'] = Na  # azimuth
        param['Nr'] = Nr  # range
        param['Ka'] = Ka  # azimuth
        param['Kr'] = Kr  # range
        param['xmin'] = xmin
        param['xmax'] = xmax
        param['ymin'] = ymin
        param['ymax'] = ymax
        param['Rnear'] = Rnear
        param['Rfar'] = Rfar
        param['tnear'] = tnear
        param['tfar'] = tfar
        param['ta'] = ta  # time array in azimuth
        param['fa'] = fa  # freq array in azimuth
        param['tr'] = tr  # time array in range
        param['fr'] = fr  # freq array in range
        param['Tsa'] = Tsa  # azimuth sampling period
        param['Tsr'] = Tsr  # range sampling period
        param['Xc'] = Xc  # X scene center
        param['Yc'] = Yc  # Y scene center
        param['Rsc'] = Rsc  # R scene center
        param['Rs0'] = Rs0  # R min scene center
        param['Ad'] = Ad  # depression angle
        param['Aon'] = Aon  # off-nadir angle
        param['As'] = As  # squint angle
        param['Ar'] = Ar  # squint angle
        param['Aba'] = Aba  # antenna elevation beamwidth
        param['Abr'] = Abr  # antenna elevation beamwidth
        param['Ai'] = Ai  # incidence angle
        param['DA'] = La / 2.0  # low squint SAR
        # slant range resolution
        param['DR'] = C / (2.0 * self._sensor['Br'])
        # ground range resolution
        param['DY'] = param['DA']  # low squint SAR
        param['DX'] = param['DR'] / (np.sin(param['Ai']) + EPS)
        param['FPa'] = FPa
        param['FPr'] = FPr
        param['BWa'] = BWa
        param['Tr'] = Tr  # range sampling times
        param['Ta'] = Ta  # azimuth sampling times
        param['t0'] = t0
        param['tac'] = tac
        param['fadc'] = fadc
        param['Bdop'] = Bdop
        param['Lsar'] = Lsar
        param['Tsar'] = Tsar
        param['Nsar'] = int(Tsar * Fsa)
        param['tstart'] = tstart
        param['tend'] = tend
        self._params = param
        logging.info('---Done!')

        return self._params

    @property
    def selection(self):
        return self._selection

    @selection.setter
    def selection(self, value):
        logging.info("---Check and set sub area selection...")
        SA = self._acquisition['SceneArea']
        SC = self._acquisition['SceneCenter']
        BC = self._acquisition['BeamCenter']
        ES = self._acquisition['EchoSize']
        Ad = self._acquisition['Ad']
        Aon = self._acquisition['Aon']
        As = self._acquisition['As']
        Ar = self._acquisition['Ar']
        Fsr = self._params['Fsr']
        Fsa = self._params['Fsa']
        Tsa = self._params['Tsa']
        Tsr = self._params['Tsr']
        Na = self._params['Na']
        Nr = self._params['Nr']
        Rnear = self._params['Rnear']
        Rfar = self._params['Rfar']
        FPr = self._params['FPr']
        H = self._sensor['H']
        Vr = self._sensor['Vr']
        Tp = self._sensor['Tp']
        Wl = self._sensor['Wl']
        La = self._sensor['La']
        GM = self._params['GeometryMode']
        Lsar = self._params['Lsar']

        if value is None:
            SubSA = SA
            SubES = ES
            SubEA = [0, 0]
        else:
            if value['SubSceneArea'] is not None:
                SubSA = value['SubSceneArea']
            else:
                SubSA = SA
            if value['SubEchoSize'] is not None:
                SubES = value['SubEchoSize']
            else:
                SubES = ES
            if value['SubEchoAnchor'] is not None:
                SubEA = value['SubEchoAnchor']
            else:
                SubEA = [0, 0]

        SubNa, SubNr = SubES
        # SubRnear = SubEA[1] * ((Rfar - Rnear) / float(Nr)) + Rnear
        # SubRfar = (SubEA[1] + SubNr) * ((Rfar - Rnear) / float(Nr)) + Rnear
        SubRnear = min_slant_range(Rnear, Fsr, SubEA[1])
        SubRfar = min_slant_range(Rnear, Fsr, SubEA[1] + SubES[1])

        SubAon = np.arccos(H / SubRnear)
        SubAbr = np.arccos(H / SubRfar) - SubAon
        SubAd = np.sign(SubAon) * (abs(SubAon) + SubAbr / 2.)
        SubFPr = np.sqrt(SubRnear**2 + SubRfar**2 - 2. * SubRnear * SubRfar * np.cos(SubAbr))

        SubTa = SubNa * Tsa
        taSub = np.linspace(-SubTa / 2.0, SubTa / 2.0, SubNa)
        faSub = np.linspace(-Fsa / 2.0, Fsa / 2.0, SubNa)

        SubRsc = 0.5 * np.sqrt(2. * SubRfar**2 + 2. * SubRnear**2 - SubFPr**2 + EPS)
        SubRs0 = SubRsc * np.cos(Ar)

        SubYsc = SubRsc * np.sin(Ar)
        SubXsc = np.sqrt(np.abs(SubRs0**2 - H**2))
        SubSC = [SubXsc, SubYsc, 0.0]

        SubRbc = H / (np.sin(Ad) + EPS)
        SubRb0 = SubRbc * np.cos(Ar)

        SubYbc = SubRbc * np.sin(Ar)
        SubXbc = np.sqrt(np.abs(SubRb0**2 - H**2))
        SubBC = [SubXbc, SubYbc, 0.0]

        SubYnear = SubRnear * np.sin(Ar)
        SubYfar = SubRfar * np.sin(Ar)
        SubXnear = np.sqrt((SubRnear ** 2 - H**2) - SubYnear ** 2 + EPS)
        SubXfar = np.sqrt((SubRfar ** 2 - H**2) - SubYfar ** 2 + EPS)
        SubSA = [SubXnear - SubXsc, SubXfar - SubXsc, -SubTa * Vr / 2.0, SubTa * Vr / 2.0]

        SubFPr0 = H / np.tan(Ad + SubAbr / 2.) - H / np.tan(Ad)
        SubFPr1 = H / np.tan(Ad - SubAbr / 2.) - H / np.tan(Ad)
        SubBA = [SubFPr0, SubFPr1, -SubTa * Vr / 2.0, SubTa * Vr / 2.0]

        tnearSub = SubRnear / C
        tfarSub = SubRfar / C
        tstartSub = 2.0 * tnearSub
        tendSub = 2.0 * tfarSub
        # tendSub = 2.0 * tfarSub + Tp

        SubTr = tendSub - tstartSub

        if GM == 'BG':
            Xc = SubBC[0]
            Yc = SubBC[1]
            SubR0 = SubRb0
            t0Sub = 2 * SubRbc / C
            trSub = np.linspace(tstartSub, tendSub, SubNr)
            # trSub = np.linspace(-SubTr / 2.0, SubTr / 2.0, SubNr) + t0Sub
            xminSub, xmaxSub, yminSub, ymaxSub = np.array(SubBA) + np.array([Xc, Xc, Yc, Yc])
        if GM == 'SG':
            Xc = SubSC[0]
            Yc = SubSC[1]
            SubR0 = SubRs0
            t0Sub = 2 * SubRsc / C
            trSub = np.linspace(tstartSub, tendSub, SubNr)
            # trSub = np.linspace(-SubTr / 2.0, SubTr / 2.0, SubNr) + t0Sub
            xminSub, xmaxSub, yminSub, ymaxSub = np.array(SubSA) + np.array([Xc, Xc, Yc, Yc])

        frSub = np.linspace(-Fsr / 2.0, Fsr / 2.0, SubNr)

        SubKa = (-2.0 * Vr**2 * np.cos(Ar)**3) / (Wl * SubRnear)
        SubLsar = Wl * SubRnear / La

        SubTsar = SubLsar / Vr
        SubBa = np.abs(SubKa * SubTsar)

        SubFPa = self._params['FPa']
        SubFPr = SubSA[1] - SubSA[0]

        self._selection = dict()
        self._selection['taSub'] = taSub
        self._selection['faSub'] = faSub
        self._selection['trSub'] = trSub
        self._selection['frSub'] = frSub
        self._selection['SubNa'] = SubNa
        self._selection['SubNr'] = SubNr
        self._selection['SubTa'] = SubTa
        self._selection['SubTr'] = SubTr
        self._selection['tnearSub'] = tnearSub
        self._selection['tfarSub'] = tfarSub
        self._selection['tstartSub'] = tstartSub
        self._selection['tendSub'] = tendSub
        self._selection['SubRnear'] = SubRnear
        self._selection['SubRfar'] = SubRfar
        self._selection['SubSceneArea'] = SubSA
        self._selection['SubBeamArea'] = SubBA
        self._selection['SubSceneCenter'] = SubSC
        self._selection['SubBeamCenter'] = SubBC
        self._selection['SubEchoAnchor'] = SubEA
        self._selection['SubEchoSize'] = SubES
        self._selection['SubFPa'] = SubFPa
        self._selection['SubFPr'] = SubFPr
        self._selection['SubRsc'] = SubRsc
        self._selection['SubRs0'] = SubRs0
        self._selection['SubRbc'] = SubRbc
        self._selection['SubRb0'] = SubRb0
        self._selection['xminSub'] = xminSub
        self._selection['xmaxSub'] = xmaxSub
        self._selection['yminSub'] = yminSub
        self._selection['ymaxSub'] = ymaxSub
        self._selection['SubAbr'] = SubAbr
        self._selection['SubAd'] = SubAd
        self._selection['SubAon'] = SubAon
        self._selection['SubKa'] = SubKa
        self._selection['SubBa'] = SubBa
        self._selection['SubTsar'] = SubTsar
        self._selection['SubLsar'] = SubLsar

        self._params['taSub'] = taSub
        self._params['faSub'] = faSub
        self._params['trSub'] = trSub
        self._params['frSub'] = frSub
        self._params['SubNa'] = SubNa
        self._params['SubNr'] = SubNr
        self._params['SubTa'] = SubTa
        self._params['SubTr'] = SubTr
        self._params['tnearSub'] = tnearSub
        self._params['tfarSub'] = tfarSub
        self._params['tstartSub'] = tstartSub
        self._params['tendSub'] = tendSub
        self._params['SubRnear'] = SubRnear
        self._params['SubRfar'] = SubRfar
        self._params['SubSceneArea'] = SubSA
        self._params['SubBeamArea'] = SubBA
        self._params['SubSceneCenter'] = SubSC
        self._params['SubBeamCenter'] = SubBC
        self._params['SubEchoAnchor'] = SubEA
        self._params['SubEchoSize'] = SubES
        self._params['SubFPa'] = SubFPa
        self._params['SubFPr'] = SubFPr
        self._params['SubRsc'] = SubRsc
        self._params['SubRs0'] = SubRs0
        self._params['SubRbc'] = SubRbc
        self._params['SubRb0'] = SubRb0
        self._params['xminSub'] = xminSub
        self._params['xmaxSub'] = xmaxSub
        self._params['yminSub'] = yminSub
        self._params['ymaxSub'] = ymaxSub
        self._params['SubAbr'] = SubAbr
        self._params['SubAd'] = SubAd
        self._params['SubAon'] = SubAon
        self._params['SubKa'] = SubKa
        self._params['SubBa'] = SubBa
        self._params['SubTsar'] = SubTsar
        self._params['SubLsar'] = SubLsar
        logging.info('---Done!')

    def printsp(self, verbose=False):
        print("===============SAR platform: ", self._name, "================")
        print("--------------------------sensor-----------------------------")
        for key in sorted(self._sensor):
            print(key, self._sensor[key])

        print("-----------------------acquisition---------------------------")
        for key in sorted(self._acquisition):
            print(key, self._acquisition[key])

        print("--------------------------params-----------------------------")
        if verbose:
            print(self._params)
        else:
            for key in sorted(self._params):
                value = self._params[key]
                if type(value) in [list, tuple, th.Tensor] and len(value) > 10:
                    print("shape of " + key + ": ", value.shape)
                else:
                    print(key, self._params[key])

        print("-------------------------selection---------------------------")
        for key in sorted(self._selection):
            value = self._selection[key]
            if type(value) in [list, tuple, th.Tensor] and len(value) > 10:
                print("shape of " + key + ": ", value.shape)
            else:
                print(key, self._selection[key])
        print("=============================================================")



def sarstore(sardata, params, file):
    """Read SAR file

    Store SAR data and platform to a file (``.pkl`` , ``.h5`` or ``.mat``)

    Args:
        sardata (dict): The SAR raw data.

        params (dict): The SAR platform for obtain the SAR data.
        file (str): Description

    Returns:
        file (str): The file path for storing SAR data and platform
    """

def sarread(file):
    """Read SAR file

    Read SAR file (``.pkl`` or ``.mat``) and returns SAR data and platform

    Args:
        file (str): SAR data (contains platform) file path.

    Returns:
        sardata (Instance SarData): The SAR raw data.

        params (Instance of SarPlat): The SAR platform for obtain the SAR data.
    """

def format_data(X, modefrom='chnl_last', modeto='chnl_first'):
    """format data

    Format data to chanel first or chanel last.

    Args:
        X (numpy array): data to be formated
        modefrom (str, optional): chnl_last  --> chanel last; chnl_first  --> chanel first
            (the default is 'chnl_last', which chanel first)
        modeto (str, optional): chnl_last  --> chanel last; chnl_first  --> chanel first
            (the default is 'chnl_last', which chanel first)

    Returns:
        X (numpy array): Formated data

    Raises:
        TypeError: X should be a 3 or 4 dimention array!
        ValueError: Unknown mode of modefrom or modeto
    """





def sls_fd(x, axis=0, wtype=None, dtype=None):
    """Sidelobe suppression in frequency domain

    Sidelobe suppression in frequency domain


    Parameters
    ----------
    x : tensor
        The input.
    axis : int, optional
        The axis for sidelobe-suppression.
    wtype : str or None, optional
        The type of window, default is None. see :func:`window`.
    dtype : torch's dtype or None, optional
        The data type. If None, use default dtype of torch.

    Returns
    -------
    tensor
        The suppressed.
    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : focusing.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import torch as th
import torchbox as tb


def focus(x, pa=None, pr=None, isfft=True, ftshift=True):
    r"""Focus image with given phase error

    Focus image in azimuth by

    .. math::
        Y(k, n_r)=\sum_{n_a=0}^{N_a-1} X(n_a, n_r) \exp \left(-j \varphi_{n_a}\right) \exp \left(-j \frac{2 \pi}{N_a} k n_a\right)

    where, :math:`\varphi_{n_a}` is the estimated azimuth phase error of the :math:`n_a`-th azimuth line, :math:`y(k, n_r)` is
    the focused pixel.

    The focus method in range is the same as azimuth.

    Args:
        x (Tensor): Defocused image data :math:`{\mathbf X} \in{\mathbb C}^{N\times N_c\times N_a\times N_r}` or
            :math:`{\mathbf X} \in{\mathbb R}^{N\times N_a\times N_r\times 2}` or
            :math:`{\mathbf X} \in{\mathbb R}^{N\times N_c\times N_a\times N_r\times 2}`.
        pa (Tensor, optional): Focus parameters in azimuth, phase error in rad unit. (the default is None, not focus)
        pr (Tensor, optional): Focus parameters in range, phase error in rad unit. (the default is None, not focus)
        isfft (bool, optional): Is need do fft (the default is True)
        ftshift (bool, optional): Is shift zero frequency to center when do fft/ifft/fftfreq (the default is True)

    Returns:
        (Tensor): A tensor of focused images.

    Raises:
        TypeError: :attr:`x` is complex and should be in complex or real represent formation!
    """

    if type(x) is not th.Tensor:
        x = th.tensor(x)

    if th.is_complex(x):
        # N, Na, Nr = x.size(0), x.size(-2), x.size(-1)
        x = th.view_as_real(x)
        crepflag = True
    elif x.size(-1) == 2:
        # N, Na, Nr = x.size(0), x.size(-3), x.size(-2)
        crepflag = False
    else:
        raise TypeError('x is complex and should be in complex or real represent formation!')

    d = x.dim()
    sizea, sizer = [1] * d, [1] * d

    if pa is not None:
        sizea[0], sizea[-3], sizea[-1] = pa.size(0), pa.size(1), 2
        epa = th.stack((th.cos(pa), -th.sin(pa)), dim=-1)
        epa = epa.reshape(sizea)

        if isfft:
            x = tb.fft(x, cdim=-1, dim=-3, keepcdim=True, shift=ftshift)
        x = tb.ematmul(x, epa, cdim=-1)
        x = tb.ifft(x, cdim=-1, dim=-3, keepcdim=True, shift=ftshift)

    if pr is not None:
        sizer[0], sizer[-2], sizer[-1] = pr.size(0), pr.size(1), 2
        epr = th.stack((th.cos(pr), -th.sin(pr)), dim=-1)
        epr = epr.reshape(sizer)

        if isfft:
            x = tb.fft(x, cdim=-1, dim=-2, keepcdim=True, shift=ftshift)
        x = tb.ematmul(x, epr, cdim=-1)
        x = tb.ifft(x, cdim=-1, dim=-2, keepcdim=True, shift=ftshift)

    if crepflag:
        x = th.view_as_complex(x)

    return x


def defocus(x, pa=None, pr=None, isfft=True, ftshift=True):
    r"""Defocus image with given phase error

    Defocus image in azimuth by

    .. math::
        Y(k, n_r)=\sum_{n_a=0}^{N_a-1} X(n_a, n_r) \exp \left(j \varphi_{n_a}\right) \exp \left(-j \frac{2 \pi}{N_a} k n_a\right)

    where, :math:`\varphi_{n_a}` is the estimated azimuth phase error of the :math:`n_a`-th azimuth line, :math:`y(k, n_r)` is
    the focused pixel.

    The defocus method in range is the same as azimuth.

    Args:
        x (Tensor): Focused image data :math:`{\mathbf X} \in{\mathbb C}^{N\times N_c\times N_a\times N_r}` or
            :math:`{\mathbf X} \in{\mathbb R}^{N\times N_a\times N_r\times 2}` or
            :math:`{\mathbf X} \in{\mathbb R}^{N\times N_c\times N_a\times N_r\times 2}`.
        pa (Tensor, optional): Defocus parameters in azimuth, phase error in rad unit. (the default is None, not focus)
        pr (Tensor, optional): Defocus parameters in range, phase error in rad unit. (the default is None, not focus)
        isfft (bool, optional): Is need do fft (the default is True)
        ftshift (bool, optional): Is shift zero frequency to center when do fft/ifft/fftfreq (the default is True)

    Returns:
        (Tensor): A tensor of defocused images.

    Raises:
        TypeError: :attr:`x` is complex and should be in complex or real represent formation!
    """

    if type(x) is not th.Tensor:
        x = th.tensor(x)

    if th.is_complex(x):
        # N, Na, Nr = x.size(0), x.size(-2), x.size(-1)
        x = th.view_as_real(x)
        crepflag = True
    elif x.size(-1) == 2:
        # N, Na, Nr = x.size(0), x.size(-3), x.size(-2)
        crepflag = False
    else:
        raise TypeError('x is complex and should be in complex or real represent formation!')

    d = x.dim()
    sizea, sizer = [1] * d, [1] * d

    if pa is not None:
        sizea[0], sizea[-3], sizea[-1] = pa.size(0), pa.size(1), 2
        epa = th.stack((th.cos(pa), th.sin(pa)), dim=-1)
        epa = epa.reshape(sizea)

        if isfft:
            x = tb.fft(x, cdim=-1, dim=-3, keepcdim=True, shift=ftshift)
        x = tb.ematmul(x, epa, cdim=-1)
        x = tb.ifft(x, cdim=-1, dim=-3, keepcdim=True, shift=ftshift)

    if pr is not None:
        sizer[0], sizer[-2], sizer[-1] = pr.size(0), pr.size(1), 2
        epr = th.stack((th.cos(pr), th.sin(pr)), dim=-1)
        epr = epr.reshape(sizer)

        if isfft:
            x = tb.fft(x, cdim=-1, dim=-2, keepcdim=True, shift=ftshift)
        x = tb.ematmul(x, epr, cdim=-1)
        x = tb.ifft(x, cdim=-1, dim=-2, keepcdim=True, shift=ftshift)

    if crepflag:
        x = th.view_as_complex(x)

    return x


if __name__ == '__main__':

    n = 8
